# -*- coding: utf-8 -*-
# @Author: Pengyao Ping
# @Date:   2023-02-16 11:01:06
# @Last Modified by:   Pengyao Ping
# @Last Modified time: 2023-09-07 00:15:00

from typing import Counter
import numpy as np
# import os
from sklearn.preprocessing import StandardScaler
from noise2read.encoding import EncodeScheme
from mpire import WorkerPool
# from tqdm import tqdm
from noise2read.utils import *
import itertools
import pickle
from noise2read.utils import MemoryMonitor

class Reads2Vectors():
    def __init__(self, logger, config, edit_dis):
        self.logger = logger
        self.edit_dis = edit_dis
        self.config = config
        # Create an instance of the MemoryMonitor
        self.MM = MemoryMonitor(self.logger)

    def read2features(self, shared_objs, idx):
        ES, ori_features = shared_objs
        cur_feature = ori_features[idx]
        features = []
        ft_fea1 = ES.descriptors("FourierTransform", cur_feature[0])
        # cg_fea1 = ES.descriptors("ChaosGame", cur_feature[0])
        entropy_fea1 = ES.descriptors("Entropy", cur_feature[0])
        fs_fea1 = ES.descriptors("FickettScore", cur_feature[0])

        features.extend(ft_fea1)
        # features.extend(cg_fea1)
        features.extend(entropy_fea1)
        features.extend(fs_fea1)

        # atomic_fea1 = ES.descriptors("atomic_number", cur_feature[0])
        # atomic_fea2 = ES.descriptors("atomic_number", cur_feature[1])
        # features.extend(atomic_fea1)
        # features.extend(atomic_fea2)

        features.extend(cur_feature[2:])
        # self.logger.debug(f'FourierTransform: {len(ft_fea1)}, ChaosGame: {len(cg_fea1)}, Entropy: {len(entropy_fea1)}, FickettScore: {len(fs_fea1)}')
        return features 

    def all_in_one_embedding(self, total_reads, genuine_df, negative_df, ambiguous_df, high_flag):
        self.logger.info("Embedding genuine and ambiguous data.")
        self.MM.start()
        if self.edit_dis == 1:
            base_lst = ['A', 'C', 'G', 'T', 'N']
            if self.config.read_type == "DNA":
                error_tyes = ["A-G", "G-A", "A-T", "T-A", "A-C", "C-A", "G-T", "T-G", "G-C", "C-G", "T-C", "C-T", "T-X", "X-T", "C-X", "X-C", "A-X", "X-A", "G-X", "X-G", "X-N", "N-X"]
                
                kmers = ['NX', 'XN', 'XA', 'XC', 'XG', 'XT', 'AX', 'CX', 'GX', 'TX'] + [''.join(i) for i in itertools.product(base_lst, repeat = 2)] + [''.join(i) for i in itertools.product(base_lst, repeat = 3)] + ['AXA', 'CXA', 'TXA', 'GXA', 'AXC', 'CXC', 'TXC', 'GXC', 'AXG', 'CXG', 'TXG', 'GXG', 'AXT', 'CXT', 'TXT', 'GXT', 'NXA', 'NXC', 'NXG', 'NXT', 'NXN','AXN','CXN','GXN','TXN']
                # kmers3 = ['ANA', 'CNA', 'TNA', 'GNA', 'ANC', 'CNC', 'TNC', 'GNC', 'ANG', 'CNG', 'TNG', 'GNG', 'ANT', 'CNT', 'TNT', 'GNT'] + [''.join(i) for i in itertools.product(base_lst, repeat = 3)]
                # kmers = kmers2 + kmers3
            elif self.config.read_type == "RNA":
                # error_tyes = ["A-G", "G-A", "A-U", "U-A", "A-C", "C-A", "G-U", "U-G", "G-C", "C-G", "U-C", "C-U", "U-N", "N-U", "C-N", "N-C", "A-N", "N-A", "G-N", "N-G"]
                error_tyes = ["A-G", "G-A", "A-T", "T-A", "A-C", "C-A", "G-T", "T-G", "G-C", "C-G", "T-C", "C-T", "T-X", "X-T", "C-X", "X-C", "A-X", "X-A", "G-X", "X-G", "X-N", "N-X"]
                kmers = ['NX', 'XN', 'XA', 'XC', 'XG', 'XT', 'AX', 'CX', 'GX', 'TX'] + [''.join(i) for i in itertools.product(base_lst, repeat = 2)] + [''.join(i) for i in itertools.product(base_lst, repeat = 3)] + ['AXA', 'CXA', 'TXA', 'GXA', 'AXC', 'CXC', 'TXC', 'GXC', 'AXG', 'CXG', 'TXG', 'GXG', 'AXT', 'CXT', 'TXT', 'GXT', 'NXC', 'NXG', 'NXU', 'NXN','AXN','CXN','GXN','UXN']

            error_tye_priors = {}
            err_tye_base_prior = 0.1
            for it in error_tyes:
                error_tye_priors[it] = err_tye_base_prior
                err_tye_base_prior += 0.01

            # base_lst = ['A', 'C', 'G', 'T']
            # kmers2 = ['AN', 'CN', 'TN', 'GN', 'NA', 'NC', 'NT', 'NG'] + [''.join(i) for i in itertools.product(base_lst, repeat = 2)]
            # kmers3 = ['ANA', 'CNA', 'TNA', 'GNA', 'ANC', 'CNC', 'TNC', 'GNC', 'ANG', 'CNG', 'TNG', 'GNG', 'ANT', 'CNT', 'TNT', 'GNT'] + [''.join(i) for i in itertools.product(base_lst, repeat = 3)]
            # kmers = kmers2 + kmers3

            self.logger.debug(f'expected kmers: {kmers}, {len(kmers)}')
            total_err_tyes = []
            total_err_kmers = []
            # for idx, row in genuine_df.iterrows():
            #     total_err_tyes.append(row['ErrorTye'])
            #     total_err_kmers.append(row['StartErrKmer'])
            #     total_err_kmers.append(row['EndErrKmer'])

            # for idx, row in ambiguous_df.iterrows():
            #     total_err_tyes.append(row['ErrorTye'])
            #     total_err_kmers.append(row['StartErrKmer'])
            #     total_err_kmers.append(row['EndErrKmer'])

            total_err_tyes.extend(genuine_df['ErrorTye'].tolist())
            total_err_tyes.extend(ambiguous_df['ErrorTye'].tolist())

            total_err_kmers.extend(genuine_df['StartErrKmer'].tolist())
            total_err_kmers.extend(ambiguous_df['StartErrKmer'].tolist())

            total_err_kmers.extend(genuine_df['EndErrKmer'].tolist())
            total_err_kmers.extend(ambiguous_df['EndErrKmer'].tolist())

            total_err_kmers_count = len(total_err_kmers)
            total_err_tyes_count = len(total_err_tyes)
            
            err_kmers2count = Counter(total_err_kmers)
            del total_err_kmers
            err_tyes2count = Counter(total_err_tyes)
            del total_err_tyes

            kmer_keys = err_kmers2count.keys()
            self.logger.debug(f'Real kmers: {kmer_keys}, {len(kmer_keys)}')
    
            not_exist_kmer = set(kmers) - set(kmer_keys)
            if not_exist_kmer:
                for mer in not_exist_kmer:
                    err_kmers2count[mer] = 0

            kmers_priors = {}
            base_prior = 0.1
            for item in kmer_keys:
                kmers_priors[item] = base_prior
                base_prior += 0.01        
            self.logger.debug(err_kmers2count)
            self.logger.debug(err_tyes2count)
            juge_indels = err_tyes2count.keys() & set(['N-X', 'X-N', 'X-A', 'X-C', 'X-G', 'X-T', 'A-X', 'C-X', 'G-X', 'T-X'])
            indel_num = len(juge_indels)
        self.MM.measure()
        genuine_feature_lst = []
        for idx, row in genuine_df.iterrows():
            # genuine_feature_lst.append(row['StartRead'])
            # genuine_feature_lst.append(row['EndRead'])
            if self.edit_dis == 1:
                cur_err_tye = row['ErrorTye']
                cur_kmer1 = row['StartErrKmer']
                cur_kmer2 = row['EndErrKmer']
                # self.logger.debug(row['StartRead'])
                # self.logger.debug(row['EndRead'])
                # self.logger.debug(f'{cur_kmer1}, {cur_kmer2}')
                cur_err_tye_val = (err_tyes2count[cur_err_tye] + error_tye_priors[cur_err_tye]) / (total_err_tyes_count + 1)
                cur_err_kmer_val1 = (err_kmers2count[cur_kmer1] + kmers_priors[cur_kmer1]) / (total_err_kmers_count + 1)
                cur_err_kmer_val2 = (err_kmers2count[cur_kmer2] + kmers_priors[cur_kmer2]) / (total_err_kmers_count + 1)
                # 
                if high_flag:
                    # genuine_feature_lst.extend([cur_err_tye_val, cur_err_kmer_val1, cur_err_kmer_val2]) #, error_tyes[row['ErrorTye']] , row["StartDegree"], row['ErrorPosition']
                    genuine_feature_lst.append((row['StartRead'], row['EndRead'], cur_err_tye_val, cur_err_kmer_val1, cur_err_kmer_val2))
                else:
                    # genuine_feature_lst.extend([row['StartReadCount'], cur_err_tye_val, cur_err_kmer_val1, cur_err_kmer_val2]) #, error_tyes[row['ErrorTye']] , row["StartDegree"], row['ErrorPosition']
                    genuine_feature_lst.append((row['StartRead'], row['EndRead'], row['StartReadCount'], cur_err_tye_val, cur_err_kmer_val1, cur_err_kmer_val2))
            else:
                # genuine_feature_lst.append(row['StartReadCount'])
                genuine_feature_lst.append((row['StartRead'], row['EndRead'], row['StartReadCount']))
            # genuine_reads_count.append(row['StartReadCount'])

        genuine_fea = self.read2vec(genuine_feature_lst)
        del genuine_feature_lst
        self.MM.measure()
        #################################################################################
        # print("encoding negative samples...")
        negative_count = 0
        neg_read2seqs = {}
        if self.edit_dis == 1:
            for idx, row in negative_df.iterrows():
                read = row['StartRead']
                if indel_num == 0:
                    pos_reads = seq2substitution(read)
                else:
                    pos_reads = enumerate_ed1_seqs(read)  
                neg_read2seqs[read] = pos_reads
                negative_count += len(pos_reads)          

        # if self.edit_dis == 2 or not self.config.over_sampling:
        # if self.edit_dis == 2 or negative_count >= 500000:
        if self.edit_dis == 2 or negative_count >= self.config.negative_sample_num:
            genuine_reads_num = len(genuine_df)
            negative_num = len(negative_df)
            if genuine_reads_num >= negative_num * 2:
                each_negative_num = int(genuine_reads_num / negative_num)
            else:
                each_negative_num = 1
        if negative_count >= self.config.negative_sample_num:
            self.logger.warning("Negative samples larger than {}, noise2read will use random sampling to generate negative samples.".format(self.config.negative_sample_num))

        negative_feature_lst = []
        for idx, row in negative_df.iterrows():
            read = row['StartRead']
            if self.edit_dis == 1:
                if negative_count < self.config.negative_sample_num:
                    pos_reads = neg_read2seqs[read]
                    for read2 in pos_reads:
                        # negative_feature_lst.append(read) 
                        # negative_feature_lst.append(read2)  
                        cur_err_tye_kmers = error_type_classification(read, read2)
                        cur_err_tye = cur_err_tye_kmers[0]
                        cur_kmer1 = cur_err_tye_kmers[1]
                        cur_kmer2 = cur_err_tye_kmers[2]
                        cur_err_tye_val = (err_tyes2count[cur_err_tye] + error_tye_priors[cur_err_tye]) / (total_err_tyes_count + 1)
                        cur_err_kmer_val1 = (err_kmers2count[cur_kmer1] + kmers_priors[cur_kmer1]) / (total_err_kmers_count + 1)
                        cur_err_kmer_val2 = (err_kmers2count[cur_kmer2] + kmers_priors[cur_kmer2]) / (total_err_kmers_count + 1)
                        if high_flag:
                            # negative_feature_lst.extend([cur_err_tye_val, cur_err_kmer_val1, cur_err_kmer_val2]) #, row["StartDegree"]
                            negative_feature_lst.append((read, read2, cur_err_tye_val, cur_err_kmer_val1, cur_err_kmer_val2))
                        else:
                            # negative_feature_lst.extend([row['StartReadCount'], cur_err_tye_val, cur_err_kmer_val1, cur_err_kmer_val2]) 
                            negative_feature_lst.append((read, read2, row['StartReadCount'], cur_err_tye_val, cur_err_kmer_val1, cur_err_kmer_val2))
                else:
                    pos_reads = neg_read2seqs[read]
                    candidates = list(pos_reads - total_reads)
                    if len(candidates) > 0:
                        for i in range(each_negative_num):
                            if len(candidates) > 0:
                                # negative_feature_lst.append(read) 
                                
                                select_read = random.choice(candidates) 
                                candidates.remove(select_read)
                                # negative_feature_lst.append(select_read)  

                                cur_err_tye_kmers = error_type_classification(read, select_read)
                                cur_err_tye = cur_err_tye_kmers[0]
                                cur_kmer1 = cur_err_tye_kmers[1]
                                cur_kmer2 = cur_err_tye_kmers[2]
                                cur_err_tye_val = (err_tyes2count[cur_err_tye] + error_tye_priors[cur_err_tye]) / (total_err_tyes_count + 1)
                                cur_err_kmer_val1 = (err_kmers2count[cur_kmer1] + kmers_priors[cur_kmer1]) / (total_err_kmers_count + 1)
                                cur_err_kmer_val2 = (err_kmers2count[cur_kmer2] + kmers_priors[cur_kmer2]) / (total_err_kmers_count + 1)
                                # 
                                if high_flag:
                                    # negative_feature_lst.extend([cur_err_tye_val, cur_err_kmer_val1, cur_err_kmer_val2]) #, row["StartDegree"]
                                    negative_feature_lst.append((read, select_read, cur_err_tye_val, cur_err_kmer_val1, cur_err_kmer_val2))
                                else:
                                    # negative_feature_lst.extend([row['StartReadCount'], cur_err_tye_val, cur_err_kmer_val1, cur_err_kmer_val2]) 
                                    negative_feature_lst.append((read, select_read, row['StartReadCount'], cur_err_tye_val, cur_err_kmer_val1, cur_err_kmer_val2))
            elif self.edit_dis == 2:
                # pos_reads = enumerate_ed2_seqs(read) 
                random_seqs = random_ed2_seq(read, total_reads, each_negative_num)
                for i in range(len(random_seqs)):
                    # negative_feature_lst.append(read) 
                    # negative_feature_lst.append(random_seqs[i])  
                    # negative_feature_lst.append(row['StartReadCount']) #, row["StartDegree"]
                    negative_feature_lst.append((read, random_seqs[i], row['StartReadCount']))

        negative_fea = self.read2vec(negative_feature_lst)
        del negative_feature_lst
        self.MM.measure()
        ###############################################################
        '''
        for idx, row in negative_df.iterrows():
            read = row['StartRead']
            if self.edit_dis == 1:
                if self.config.over_sampling:
                    if indel_num == 0:
                        pos_reads = seq2substitution(read)
                    else:
                        pos_reads = enumerate_ed1_seqs(read)
                    # candidates = list(pos_reads - total_reads)
                    # for i in range(each_negative_num):
                    for read2 in pos_reads:
                        negtive_reads_lst1.append(read) 
                        # select_read = random.choice(candidates) 
                        # candidates.remove(select_read)
                        negtive_reads_lst2.append(read2)  

                        cur_err_tye_kmers = error_type_classification(read, read2)
                        cur_err_tye = cur_err_tye_kmers[0]
                        cur_kmer1 = cur_err_tye_kmers[1]
                        cur_kmer2 = cur_err_tye_kmers[2]
                        cur_err_tye_val = (err_tyes2count[cur_err_tye] + error_tye_priors[cur_err_tye]) / (total_err_tyes_count + 1)
                        cur_err_kmer_val1 = (err_kmers2count[cur_kmer1] + kmers_priors[cur_kmer1]) / (total_err_kmers_count + 1)
                        cur_err_kmer_val2 = (err_kmers2count[cur_kmer2] + kmers_priors[cur_kmer2]) / (total_err_kmers_count + 1)
                        # 
                        if high_flag:
                            negtive_reads_features.append([cur_err_tye_val, cur_err_kmer_val1, cur_err_kmer_val2]) #, row["StartDegree"]
                        else:
                            negtive_reads_features.append([row['StartReadCount'], cur_err_tye_val, cur_err_kmer_val1, cur_err_kmer_val2]) 
                else:
                    if indel_num == 0:
                        pos_reads = seq2substitution(read)
                    else:
                        pos_reads = enumerate_ed1_seqs(read)
                    candidates = list(pos_reads - total_reads)
                    for i in range(each_negative_num):
                        negtive_reads_lst1.append(read) 
                        select_read = random.choice(candidates) 
                        candidates.remove(select_read)
                        negtive_reads_lst2.append(select_read)  

                        cur_err_tye_kmers = error_type_classification(read, select_read)
                        cur_err_tye = cur_err_tye_kmers[0]
                        cur_kmer1 = cur_err_tye_kmers[1]
                        cur_kmer2 = cur_err_tye_kmers[2]
                        cur_err_tye_val = (err_tyes2count[cur_err_tye] + error_tye_priors[cur_err_tye]) / (total_err_tyes_count + 1)
                        cur_err_kmer_val1 = (err_kmers2count[cur_kmer1] + kmers_priors[cur_kmer1]) / (total_err_kmers_count + 1)
                        cur_err_kmer_val2 = (err_kmers2count[cur_kmer2] + kmers_priors[cur_kmer2]) / (total_err_kmers_count + 1)
                        # 
                        if high_flag:
                            negtive_reads_features.append([cur_err_tye_val, cur_err_kmer_val1, cur_err_kmer_val2]) #, row["StartDegree"]
                        else:
                            negtive_reads_features.append([row['StartReadCount'], cur_err_tye_val, cur_err_kmer_val1, cur_err_kmer_val2]) 
            elif self.edit_dis == 2:
                # pos_reads = enumerate_ed2_seqs(read) 
                random_seqs = random_ed2_seq(read, total_reads, each_negative_num)
                for i in range(len(random_seqs)):
                    negtive_reads_lst1.append(read) 
                    negtive_reads_lst2.append(random_seqs[i])  
                    negtive_reads_features.append([row['StartReadCount']]) #, row["StartDegree"]
                    # negtive_reads_count.append(row['StartReadCount']) 
        '''
        ambiguous_feature_lst = []
        for idx, row in ambiguous_df.iterrows():
            if self.edit_dis == 1:
                cur_err_tye = row['ErrorTye']
                cur_kmer1 = row['StartErrKmer']
                cur_kmer2 = row['EndErrKmer']
                cur_err_tye_val = (err_tyes2count[cur_err_tye] + error_tye_priors[cur_err_tye]) / (total_err_tyes_count + 1)
                cur_err_kmer_val1 = (err_kmers2count[cur_kmer1] + kmers_priors[cur_kmer1]) / (total_err_kmers_count + 1)
                cur_err_kmer_val2 = (err_kmers2count[cur_kmer2] + kmers_priors[cur_kmer2]) / (total_err_kmers_count + 1)
                # 
                if high_flag:
                    ambiguous_feature_lst.append((row['StartRead'], row['EndRead'], cur_err_tye_val, cur_err_kmer_val1, cur_err_kmer_val2))
                else:
                    ambiguous_feature_lst.append((row['StartRead'], row['EndRead'], row['StartReadCount'], cur_err_tye_val, cur_err_kmer_val1, cur_err_kmer_val2))
            else:
                ambiguous_feature_lst.append((row['StartRead'], row['EndRead'], row['StartReadCount']))

        ambiguous_fea = self.read2vec(ambiguous_feature_lst)
        del ambiguous_feature_lst
        self.MM.measure()
        ##################################################################
        read_features = genuine_fea + negative_fea
        feature_len = len(read_features[0])
        train_data = np.array(read_features, dtype=object)
        gen_len = len(genuine_fea)
        neg_len = len(negative_fea)
        del read_features, genuine_fea, negative_fea

        labels = np.array([1] * gen_len + [0] * neg_len)
        shape1 = ((gen_len + neg_len), feature_len)
        train_data.reshape(shape1)   

        ambiguous_data = np.array(ambiguous_fea, dtype=object)
        shape2 = (len(ambiguous_fea), len(ambiguous_fea[0]))
        del ambiguous_fea
        ambiguous_data.reshape(shape2) 
        # scaling data
        self.logger.debug(train_data.shape)
        self.logger.debug(ambiguous_data.shape)
        if self.edit_dis == 1:
            train, ambiguous = self.scaler(train_data, ambiguous_data, high_flag)
        else:
            train, ambiguous = self.scaler2(train_data, ambiguous_data)
        self.logger.debug(train[0])
        del train_data, ambiguous_data
        self.MM.measure()
        self.MM.stop()
        return train, labels, ambiguous

    def scaler2(self, lab_fea, ambiguous_ulab_fea):
        # pos0 = 179
        # pos0 = 65
        pos0 = 27
        lab_fea0 = lab_fea[:,0:pos0]

        ulab_fea0 = ambiguous_ulab_fea[:,0:pos0]
        scaler0 = StandardScaler()
        lab_scale_f0 = scaler0.fit_transform(lab_fea0)
        ulab_scale_f0 = scaler0.transform(ulab_fea0)

        # # read integer encoding
        # pos = (self.config.read_max_len+1) * 2
        # # pos = (self.config.read_max_len+1) * 4  * 2
        # lab_fea1 = lab_fea[:,pos0:pos+pos0]

        # ulab_fea1 = ambiguous_ulab_fea[:,pos0:pos+pos0]
        # scaler1 = StandardScaler()
        # lab_scale_f1 = scaler1.fit_transform(lab_fea1)
        # ulab_scale_f1 = scaler1.transform(ulab_fea1)

        # read count
        lab_fea2 = lab_fea[:,-4].reshape(-1, 1)
        ulab_fea2 = ambiguous_ulab_fea[:,-4].reshape(-1, 1)
        scaler2 = StandardScaler()
        lab_scale_f2 = scaler2.fit_transform(lab_fea2)
        ulab_scale_f2 = scaler2.transform(ulab_fea2)

        # lab_scale_f = np.hstack((lab_scale_f0, lab_scale_f1, lab_scale_f2))
        # ulab_scale_f = np.hstack((ulab_scale_f0, ulab_scale_f1, ulab_scale_f2))
        # del lab_scale_f0, lab_scale_f1, lab_scale_f2, ulab_scale_f0, ulab_scale_f1, ulab_scale_f2
        lab_scale_f = np.hstack((lab_scale_f0, lab_scale_f2))
        ulab_scale_f = np.hstack((ulab_scale_f0, ulab_scale_f2))
        del lab_scale_f0, lab_scale_f2, ulab_scale_f0, ulab_scale_f2
        return lab_scale_f, ulab_scale_f 

    def scaler(self, lab_fea, ambiguous_ulab_fea, high_flag):
        # pos0 = 179
        # pos0 = 65
        pos0 = 27
        lab_fea0 = lab_fea[:,0:pos0]

        ulab_fea0 = ambiguous_ulab_fea[:,0:pos0]
        scaler0 = StandardScaler()
        lab_scale_f0 = scaler0.fit_transform(lab_fea0)
        ulab_scale_f0 = scaler0.transform(ulab_fea0)

        # read integer encoding
        # pos = (self.config.read_max_len+1) * 2
        # # pos = self.config.read_max_len+1
        # # pos = (self.config.read_max_len+1) * 4 * 2
        # lab_fea1 = lab_fea[:,pos0:pos+pos0]

        # ulab_fea1 = ambiguous_ulab_fea[:,pos0:pos+pos0]
        # scaler1 = StandardScaler()
        # lab_scale_f1 = scaler1.fit_transform(lab_fea1)
        # ulab_scale_f1 = scaler1.transform(ulab_fea1)

        # error type and error kmers
        lab_fea3 = lab_fea[:,-3:]
        ulab_fea3 = ambiguous_ulab_fea[:,-3:]
        scaler3 = StandardScaler()
        lab_scale_f3 = scaler3.fit_transform(lab_fea3)
        ulab_scale_f3 = scaler3.transform(ulab_fea3)
        
        # read count
        if not high_flag:
            lab_fea2 = lab_fea[:,-4].reshape(-1, 1)
            ulab_fea2 = ambiguous_ulab_fea[:,-4].reshape(-1, 1)
            scaler2 = StandardScaler()
            lab_scale_f2 = scaler2.fit_transform(lab_fea2)
            ulab_scale_f2 = scaler2.transform(ulab_fea2)

            # lab_scale_f = np.hstack((lab_scale_f0, lab_scale_f1, lab_scale_f2, lab_scale_f3))
            # ulab_scale_f = np.hstack((ulab_scale_f0, ulab_scale_f1, ulab_scale_f2, ulab_scale_f3))
            # # lab_scale_f = np.hstack((lab_scale_f0, lab_scale_f1, lab_scale_f3))
            # # ulab_scale_f = np.hstack((ulab_scale_f0, ulab_scale_f1, ulab_scale_f3))
            # del lab_scale_f0, lab_scale_f1, lab_scale_f2, lab_scale_f3, ulab_scale_f0, ulab_scale_f1, ulab_scale_f2, ulab_scale_f3

            lab_scale_f = np.hstack((lab_scale_f0, lab_scale_f2, lab_scale_f3))
            ulab_scale_f = np.hstack((ulab_scale_f0, ulab_scale_f2, ulab_scale_f3))
            # lab_scale_f = np.hstack((lab_scale_f0, lab_scale_f1, lab_scale_f3))
            # ulab_scale_f = np.hstack((ulab_scale_f0, ulab_scale_f1, ulab_scale_f3))
            del lab_scale_f0, lab_scale_f2, lab_scale_f3, ulab_scale_f0, ulab_scale_f2, ulab_scale_f3

        else:
            # lab_scale_f = np.hstack((lab_scale_f0, lab_scale_f1, lab_scale_f3))
            # ulab_scale_f = np.hstack((ulab_scale_f0, ulab_scale_f1, ulab_scale_f3))
            # del lab_scale_f0, lab_scale_f1, lab_scale_f3, ulab_scale_f0, ulab_scale_f1, ulab_scale_f3

            lab_scale_f = np.hstack((lab_scale_f0, lab_scale_f3))
            ulab_scale_f = np.hstack((ulab_scale_f0, ulab_scale_f3))
            del lab_scale_f0, lab_scale_f3, ulab_scale_f0, ulab_scale_f3

        # lab_scale_f = np.hstack((lab_scale_f0, lab_scale_f1, lab_scale_f2, lab_scale_f3))
        # ulab_scale_f = np.hstack((ulab_scale_f0, ulab_scale_f1, ulab_scale_f2, ulab_scale_f3))
        
        return lab_scale_f, ulab_scale_f 

    '''
    def read2features(self, ES, ori_feature):
        # ES, reads_lst1, reads_lst2, other_features = shared_objects
        features = []
        
        ###########################################################################
        ft_fea1 = ES.descriptors("FourierTransform", ori_feature[0])
        cg_fea1 = ES.descriptors("ChaosGame", ori_feature[0])
        entropy_fea1 = ES.descriptors("Entropy", ori_feature[0])
        fs_fea1 = ES.descriptors("FickettScore", ori_feature[0])
        features.extend(ft_fea1)
        features.extend(cg_fea1)
        features.extend(entropy_fea1)
        features.extend(fs_fea1)
        atomic_fea1 = ES.descriptors("atomic_number",ori_feature[0])
        atomic_fea2 = ES.descriptors("atomic_number", ori_feature[1])
        features.extend(atomic_fea1)
        features.extend(atomic_fea2)
        
        features.extend(ori_feature[2:])
        return features 
    '''
    def read2vec(self, original_features_lst):  
        ES = EncodeScheme(self.config.read_max_len, self.config.entropy_kmer, self.config.entropy_q, self.config.kmer_freq, self.config.read_type)

        if len(original_features_lst) > (self.config.chunks_num * 10):
            chunk_size = len(original_features_lst) // self.config.chunks_num
            # remainder = len(original_features_lst) % self.config.chunks_num
            # chunks = [original_features_lst[i:i+chunk_size] for i in range(0, len(original_features_lst), chunk_size)]
            combined_data = []
            if chunk_size >= self.config.num_workers:
                chunks = [original_features_lst[i:i+chunk_size] for i in range(0, len(original_features_lst), chunk_size)]
                chunk_names = []
                # combined_data = []
                # for i in range(len(chunks)):
                i = 0
                while chunks:
                    chunk = chunks.pop(0)
                    vectors = []
                    try:
                        shared_objects = ES, chunk
                        with WorkerPool(self.config.num_workers, shared_objects=shared_objects, start_method='fork') as pool:
                            for item in pool.imap(self.read2features, range(len(chunk))):
                                vectors.append(item)
                        del shared_objects
                    except KeyboardInterrupt:
                        # Handle termination signal (Ctrl+C)
                        pool.terminate()  # Terminate the WorkerPool before exiting
                    except Exception:
                        # Handle other exceptions
                        pool.terminate()  # Terminate the WorkerPool before exiting
                        raise
                    del chunk        
                    # Generate the pickle file name
                    file_name = self.config.result_dir + f"chunk_{i}.pickle"
                    i += 1
                    # Write the vectors to the pickle file
                    with open(file_name, "wb") as file:
                        pickle.dump(vectors, file)
                    chunk_names.append(file_name)
                    del vectors
                del chunks
                
                for file_name in chunk_names:
                    with open(file_name, "rb") as file:
                        vectors = pickle.load(file)
                        combined_data.extend(vectors)
                        del vectors
                    os.remove(file_name)
            else:
                try:
                    shared_objects = ES, original_features_lst
                    with WorkerPool(self.config.num_workers, shared_objects=shared_objects, start_method='fork') as pool:
                        for item in pool.imap(self.read2features, range(len(original_features_lst))):
                            combined_data.append(item)
                    del shared_objects
                except KeyboardInterrupt:
                    # Handle termination signal (Ctrl+C)
                    pool.terminate()  # Terminate the WorkerPool before exiting
                except Exception:
                    # Handle other exceptions
                    pool.terminate()  # Terminate the WorkerPool before exiting
                    raise   
            del ES, original_features_lst         
            return combined_data
        else:
            try:
                vectors = []
                shared_objects = ES, original_features_lst
                with WorkerPool(self.config.num_workers, shared_objects=shared_objects, start_method='fork') as pool:
                    for item in pool.imap(self.read2features, range(len(original_features_lst))):
                        vectors.append(item)
                del shared_objects
            except KeyboardInterrupt:
                # Handle termination signal (Ctrl+C)
                pool.terminate()  # Terminate the WorkerPool before exiting
            except Exception:
                # Handle other exceptions
                pool.terminate()  # Terminate the WorkerPool before exiting
                raise
            del ES, original_features_lst    
            return vectors

    
    def high_all_in_one_embedding(self, genuine_df, negative_df, new_negative_df, ambiguous_df):
        self.logger.info("Embedding genuine and high ambiguous data.")
        self.MM.start()
        base_lst = ['A', 'C', 'G', 'T', 'N']
        if self.config.read_type == "DNA":
            error_tyes = ["A-G", "G-A", "A-T", "T-A", "A-C", "C-A", "G-T", "T-G", "G-C", "C-G", "T-C", "C-T", "T-X", "X-T", "C-X", "X-C", "A-X", "X-A", "G-X", "X-G", "X-N", "N-X", 'A-N', 'T-N','G-N','C-N','N-A','N-T', 'N-C', 'N-G']
            kmers = ['NX', 'XN', 'XA', 'XC', 'XG', 'XT', 'AX', 'CX', 'GX', 'TX'] + [''.join(i) for i in itertools.product(base_lst, repeat = 2)] + [''.join(i) for i in itertools.product(base_lst, repeat = 3)] + ['AXA', 'CXA', 'TXA', 'GXA', 'AXC', 'CXC', 'TXC', 'GXC', 'AXG', 'CXG', 'TXG', 'GXG', 'AXT', 'CXT', 'TXT', 'GXT', 'NXA', 'NXC', 'NXG', 'NXT', 'NXN','AXN','CXN','GXN','TXN']
            # kmers3 = ['ANA', 'CNA', 'TNA', 'GNA', 'ANC', 'CNC', 'TNC', 'GNC', 'ANG', 'CNG', 'TNG', 'GNG', 'ANT', 'CNT', 'TNT', 'GNT'] + [''.join(i) for i in itertools.product(base_lst, repeat = 3)]
            # kmers = kmers2 + kmers3
        elif self.config.read_type == "RNA":
            # error_tyes = ["A-G", "G-A", "A-U", "U-A", "A-C", "C-A", "G-U", "U-G", "G-C", "C-G", "U-C", "C-U", "U-N", "N-U", "C-N", "N-C", "A-N", "N-A", "G-N", "N-G"]
            error_tyes = ["A-G", "G-A", "A-T", "T-A", "A-C", "C-A", "G-T", "T-G", "G-C", "C-G", "T-C", "C-T", "T-X", "X-T", "C-X", "X-C", "A-X", "X-A", "G-X", "X-G", "X-N", "N-X", 'A-N', 'T-N','G-N','C-N','N-A','N-T', 'N-C', 'N-G']
            kmers = ['NX', 'XN', 'XA', 'XC', 'XG', 'XT', 'AX', 'CX', 'GX', 'TX'] + [''.join(i) for i in itertools.product(base_lst, repeat = 2)] + [''.join(i) for i in itertools.product(base_lst, repeat = 3)] + ['AXA', 'CXA', 'TXA', 'GXA', 'AXC', 'CXC', 'TXC', 'GXC', 'AXG', 'CXG', 'TXG', 'GXG', 'AXT', 'CXT', 'TXT', 'GXT', 'NXA', 'NXC', 'NXG', 'NXU', 'NXN','AXN','CXN','GXN','UXN']

        error_tye_priors = {}
        err_tye_base_prior = 0.1
        for it in error_tyes:
            error_tye_priors[it] = err_tye_base_prior
            err_tye_base_prior += 0.01

        self.logger.debug(len(kmers))
        total_err_tyes = []
        total_err_kmers = []
        total_err_tyes.extend(genuine_df['ErrorTye'].tolist())
        total_err_tyes.extend(ambiguous_df['ErrorTye'].tolist())
        total_err_tyes.extend(new_negative_df['ErrorTye'].tolist())

        total_err_kmers.extend(genuine_df['StartErrKmer'].tolist())
        total_err_kmers.extend(ambiguous_df['StartErrKmer'].tolist())
        total_err_kmers.extend(new_negative_df['StartErrKmer'].tolist())

        total_err_kmers.extend(genuine_df['EndErrKmer'].tolist())
        total_err_kmers.extend(ambiguous_df['EndErrKmer'].tolist())
        total_err_kmers.extend(new_negative_df['EndErrKmer'].tolist())
        self.MM.measure()
        # for idx, row in genuine_df.iterrows():
        #     total_err_tyes.append(row['ErrorTye'])
        #     total_err_kmers.append(row['StartErrKmer'])
        #     total_err_kmers.append(row['EndErrKmer'])

        # for idx, row in ambiguous_df.iterrows():
        #     total_err_tyes.append(row['ErrorTye'])
        #     total_err_kmers.append(row['StartErrKmer'])
        #     total_err_kmers.append(row['EndErrKmer'])

        # for idx, row in new_negative_df.iterrows():
        #     # read = row['StartRead']
        #     total_err_tyes.append(row['ErrorTye'])
        #     total_err_kmers.append(row['StartErrKmer'])
        #     total_err_kmers.append(row['EndErrKmer'])
            
        total_err_kmers_count = len(total_err_kmers)
        total_err_tyes_count = len(total_err_tyes)
        
        err_kmers2count = Counter(total_err_kmers)
        del total_err_kmers
        err_tyes2count = Counter(total_err_tyes)
        del total_err_tyes
        self.logger.debug(len(err_kmers2count.keys()))
        kmer_keys = err_kmers2count.keys()

        not_exist_kmer = set(kmers) - set(kmer_keys)
        if not_exist_kmer:
            for mer in not_exist_kmer:
                err_kmers2count[mer] = 0

        kmers_priors = {}
        base_prior = 0.1
        for item in kmer_keys:
            kmers_priors[item] = base_prior
            base_prior += 0.01        
        self.logger.debug(err_kmers2count)
        self.logger.debug(err_tyes2count)
        self.MM.measure()
        ##################################################################################
        genuine_feature_lst = []
        for idx, row in genuine_df.iterrows():
            cur_err_tye = row['ErrorTye']
            cur_kmer1 = row['StartErrKmer']
            cur_kmer2 = row['EndErrKmer']
            # self.logger.debug(row['StartRead'])
            # self.logger.debug(row['EndRead'])
            # self.logger.debug(f'{cur_kmer1}, {cur_kmer2}')
            cur_err_tye_val = (err_tyes2count[cur_err_tye] + error_tye_priors[cur_err_tye]) / (total_err_tyes_count + 1)
            cur_err_kmer_val1 = (err_kmers2count[cur_kmer1] + kmers_priors[cur_kmer1]) / (total_err_kmers_count + 1)
            cur_err_kmer_val2 = (err_kmers2count[cur_kmer2] + kmers_priors[cur_kmer2]) / (total_err_kmers_count + 1)
            # genuine_feature_lst.extend([cur_err_tye_val, cur_err_kmer_val1, cur_err_kmer_val2]) #, error_tyes[row['ErrorTye']] , row["StartDegree"], row['ErrorPosition']
            genuine_feature_lst.append((row['StartRead'], row['EndRead'], cur_err_tye_val, cur_err_kmer_val1, cur_err_kmer_val2))
        genuine_fea = self.read2vec(genuine_feature_lst)
        del genuine_feature_lst
        self.MM.measure()
        #################################################################################
        negative_feature_lst = []
        for idx, row in new_negative_df.iterrows():
            # negative_feature_lst.append()
            # negative_feature_lst.append()
            cur_err_tye = row['ErrorTye']
            cur_kmer1 = row['StartErrKmer']
            cur_kmer2 = row['EndErrKmer']
            cur_err_tye_val = (err_tyes2count[cur_err_tye] + error_tye_priors[cur_err_tye]) / (total_err_tyes_count + 1)
            cur_err_kmer_val1 = (err_kmers2count[cur_kmer1] + kmers_priors[cur_kmer1]) / (total_err_kmers_count + 1)
            cur_err_kmer_val2 = (err_kmers2count[cur_kmer2] + kmers_priors[cur_kmer2]) / (total_err_kmers_count + 1)
            # negative_feature_lst.extend([cur_err_tye_val, cur_err_kmer_val1, cur_err_kmer_val2]) #, row["StartDegree"]
            negative_feature_lst.append((row['StartRead'], row['EndRead'], cur_err_tye_val, cur_err_kmer_val1, cur_err_kmer_val2))
        # isolates negative
        if self.config.iso_neg_high:
            for idx, row in negative_df.iterrows():
                read = row['StartRead']
                pos_reads = enumerate_ed1_seqs(read)
                for read2 in pos_reads:
                    # negative_feature_lst.append(read) 
                    # negative_feature_lst.append(read2)  
                    cur_err_tye_kmers = error_type_classification(read, read2)
                    cur_err_tye = cur_err_tye_kmers[0]
                    cur_kmer1 = cur_err_tye_kmers[1]
                    cur_kmer2 = cur_err_tye_kmers[2]
                    cur_err_tye_val = (err_tyes2count[cur_err_tye] + error_tye_priors[cur_err_tye]) / (total_err_tyes_count + 1)
                    cur_err_kmer_val1 = (err_kmers2count[cur_kmer1] + kmers_priors[cur_kmer1]) / (total_err_kmers_count + 1)
                    cur_err_kmer_val2 = (err_kmers2count[cur_kmer2] + kmers_priors[cur_kmer2]) / (total_err_kmers_count + 1)
                    # negative_feature_lst.extend([cur_err_tye_val, cur_err_kmer_val1, cur_err_kmer_val2])  
                    negative_feature_lst.append((read, read2, cur_err_tye_val, cur_err_kmer_val1, cur_err_kmer_val2))
        
        negative_fea = self.read2vec(negative_feature_lst)
        del negative_feature_lst
        self.MM.measure()
        ###############################################################
        ambiguous_feature_lst = []
        for idx, row in ambiguous_df.iterrows():
            # ambiguous_feature_lst.append(row['StartRead'])
            # ambiguous_feature_lst.append(row['EndRead'])
            cur_err_tye = row['ErrorTye']
            cur_kmer1 = row['StartErrKmer']
            cur_kmer2 = row['EndErrKmer']
            cur_err_tye_val = (err_tyes2count[cur_err_tye] + error_tye_priors[cur_err_tye]) / (total_err_tyes_count + 1)
            cur_err_kmer_val1 = (err_kmers2count[cur_kmer1] + kmers_priors[cur_kmer1]) / (total_err_kmers_count + 1)
            cur_err_kmer_val2 = (err_kmers2count[cur_kmer2] + kmers_priors[cur_kmer2]) / (total_err_kmers_count + 1)
            # ambiguous_feature_lst.extend([cur_err_tye_val, cur_err_kmer_val1, cur_err_kmer_val2])#, row["StartDegree"]
            ambiguous_feature_lst.append((row['StartRead'], row['EndRead'], cur_err_tye_val, cur_err_kmer_val1, cur_err_kmer_val2))
        ambiguous_fea = self.read2vec(ambiguous_feature_lst)
        del ambiguous_feature_lst
        self.MM.measure()
        ##################################################################
        read_features = genuine_fea + negative_fea
        labels = np.array([1] * len(genuine_fea) + [0] * len(negative_fea))
        train_data = np.array(read_features)
        shape1 = (len(labels), len(read_features[0]))
        train_data.reshape(shape1)   

        ambiguous_data = np.array(ambiguous_fea)
        shape2 = (len(ambiguous_fea), len(ambiguous_fea[0]))
        ambiguous_data.reshape(shape2) 
        # scaling data
        self.logger.debug(train_data.shape)
        self.logger.debug(ambiguous_data.shape)
        train, ambiguous = self.scaler(train_data, ambiguous_data, high_flag=True)
        self.logger.debug(train[0])
        del train_data, ambiguous_data, genuine_fea, negative_fea, ambiguous_fea, read_features
        self.MM.measure()
        self.MM.stop()
        return train, labels, ambiguous

    '''
    def read2vec(self, reads_lst1, reads_lst2, other_features):
        read_features = []    
        ES = EncodeScheme(self.config.read_max_len, self.config.entropy_kmer, self.config.entropy_q, self.config.kmer_freq, self.config.read_type)

        shared_objects = ES, reads_lst1, reads_lst2, other_features
        # with WorkerPool(self.config.num_workers, shared_objects=shared_objects, start_method='fork') as pool:
        #     with tqdm(total=len(reads_lst1), desc=self.logger.info("Encoding reads")) as pbar:   
        #         for fea_lst in pool.imap(self.read2features, range(len(reads_lst1))):
        #             read_features.append(fea_lst)
        #             pbar.update()  
        with WorkerPool(self.config.num_workers, shared_objects=shared_objects, start_method='fork') as pool:
            for fea_lst in pool.imap(self.read2features, range(len(reads_lst1)), progress_bar=self.config.verbose):
                read_features.append(fea_lst)
        
        self.logger.debug(f'{len(read_features)}, {len(read_features[0])}')
        return read_features
    '''

    '''
    def read2features(self, shared_objects, i):
        ES, reads_lst1, reads_lst2, other_features = shared_objects
        features = []
        # pd_fe = ES.descriptors("PairDistance", reads_lst[i])
        # features.extend(pd_fe)    
        ###########################################################################
        # features.append(read_count_lst[i]) 
        # features.append(err_pos_lst[i]) 
        # freq_pos.append([ori_seq_freq, err_pos])              
        ###########################################################################
        ft_fea1 = ES.descriptors("FourierTransform", reads_lst1[i])
        cg_fea1 = ES.descriptors("ChaosGame", reads_lst1[i])
        entropy_fea1 = ES.descriptors("Entropy", reads_lst1[i])
        fs_fea1 = ES.descriptors("FickettScore", reads_lst1[i])

        features.extend(ft_fea1)
        features.extend(cg_fea1)
        features.extend(entropy_fea1)
        features.extend(fs_fea1)
        # print(len(features))
        # ft_fea2 = ES.descriptors("FourierTransform", reads_lst2[i])
        # cg_fea2 = ES.descriptors("ChaosGame", reads_lst2[i])
        # # entropy_fea2 = ES.descriptors("Entropy", reads_lst2[i])
        # fs_fea2 = ES.descriptors("FickettScore", reads_lst2[i])

        # features.extend(ft_fea2)
        # features.extend(cg_fea2)
        # # features.extend(entropy_fea2)
        # features.extend(fs_fea2)
        # self.logger.debug(len(features))
        atomic_fea1 = ES.descriptors("atomic_number", reads_lst1[i])
        atomic_fea2 = ES.descriptors("atomic_number", reads_lst2[i])
        # atomic_fea1 = ES.descriptors("binary", reads_lst1[i])
        # atomic_fea2 = ES.descriptors("binary", reads_lst2[i])
        # print(int(other_features[i][0]))
        features.extend(atomic_fea1)
        features.extend(atomic_fea2)
        # onehot_fea = ES.descriptors("OneHot", reads_lst1[i])
        # features.extend(onehot_fea)
        features.extend(other_features[i])
        # self.logger.debug(f'FourierTransform: {len(ft_fea)}, ChaosGame: {len(cg_fea)}, Entropy: {len(entropy_fea)}, FickettScore: {len(fs_fea)}')
        return features 
    '''

   


# -*- coding: utf-8 -*-
# @Author: Pengyao Ping
# @Date:   2023-02-16 11:01:06
# @Last Modified by:   Pengyao Ping
# @Last Modified time: 2023-09-07 16:59:49

import collections
from Bio import SeqIO
from Bio.SeqRecord import SeqRecord
import os
from noise2read.utils import *
from noise2read.data_generation import DataGneration
from noise2read.error_orrection import ErrorCorrection
# from noise2read.data_analysis import DataAnalysis
import editdistance
from tqdm import tqdm
from Bio.Seq import Seq
import json

class DataProcessing():
    # def __init__(self, logger, num_workers, output_dir, umi_start=None, umi_end=None, non_umi_start=None):
    def __init__(self, logger, config):
        self.logger = logger
        self.config = config
        
        if os.path.exists(self.config.result_dir):
            self.logger.info("Directory '%s' already exists" % self.config.result_dir)
        else:
            os.makedirs(self.config.result_dir)
        if os.path.exists(self.config.result_dir + 'raw/'):
            self.logger.info(f"Directory {self.config.result_dir}raw/ already exists")
        else:
            os.makedirs(self.config.result_dir + 'raw/')       
        if os.path.exists(self.config.result_dir + 'true/'):
            self.logger.info(f"Directory {self.config.result_dir}true/ already exists")
        else:
            os.makedirs(self.config.result_dir + 'true/') 
               
    def extract_umis(self, original_data, umi_start, umi_end, non_umi_start):
        record_iterator, ff_type = parse_data(original_data) 
        self.logger.info("Read Data")
        umi_records = []
        non_umi_records = []

        for item in tqdm(record_iterator):
            seq = str(item.seq)
            if 'N' not in seq:
                umi_qual = {}
                non_umi_qual = {}
                if ff_type == 'fastq' or ff_type == 'fq' or ff_type == 'fastq.gz' or ff_type == 'fq.gz':
                    val = item.letter_annotations['phred_quality']            
                umi_desc = "length=" + str(umi_end-umi_start)
                non_umi_desc = "length=" + str(len(seq) - non_umi_start)
                if ff_type == "fastq": 
                    umi_qual['phred_quality'] = val[umi_start:umi_end]
                    non_umi_qual['phred_quality'] = val[non_umi_start:]
                    umi_rec = SeqRecord(Seq(seq[umi_start:umi_end]), id=item.id, description=umi_desc, letter_annotations=umi_qual)   
                    non_umi_rec = SeqRecord(Seq(seq[non_umi_start:]), id=item.id, description=non_umi_desc, letter_annotations=non_umi_qual)   
                elif ff_type == "fasta":
                    umi_rec = SeqRecord(Seq(seq[umi_start:umi_end]), id=item.id, description=umi_desc)        
                    non_umi_rec = SeqRecord(Seq(seq[non_umi_start:]), id=item.id, description=non_umi_desc)  
                umi_records.append(umi_rec)
                non_umi_records.append(non_umi_rec)        
        self.logger.info("Rewrite umi and non-umi Data")
        raw_base_name = original_data.split('/')[-1]
        umi_raw_dataset = self.config.result_dir + "umi_" + raw_base_name
        non_umi_raw_dataset = self.config.result_dir + "umi_in_name_" + raw_base_name
        with open(umi_raw_dataset, "w") as handle:
            SeqIO.write(umi_records, handle, ff_type)        
        with open(non_umi_raw_dataset, "w") as handle:
            SeqIO.write(non_umi_records, handle, ff_type) 
        self.logger.info("Split umi and non-umis completed.") 
        return umi_raw_dataset, non_umi_raw_dataset

    def real_umi_data(self, original_data):
        # step1: extract umis to a sperate fastq file
        umi_raw_dataset, non_umi_raw_dataset = self.extract_umis(original_data, self.config.umi_start, self.config.umi_end, self.config.non_umi_start)
        # step2: correct umi errors
        # self.config.input_file = umi_raw_dataset
        # DG = DataGneration(self.logger, self.config)

        # genuine_df = DG.extract_umi_genuine_errs(umi_raw_dataset)
        # # ##############################################################
        # EC = ErrorCorrection(self.logger, self.config)
        # corrected_file = EC.umi_correction(umi_raw_dataset, genuine_df)

        # base_name = umi_raw_dataset.split("/")[-1].split(".fastq")[0]
        # step3: use umi and real dataset to generate raw and true dataset
        # raw_file, true_file = self.raw_true_umi(corrected_file, non_umi_raw_dataset)
        raw_file, true_file = self.raw_true_umi(umi_raw_dataset, non_umi_raw_dataset)
        return raw_file, true_file

    def raw_true_umi(self, correct_umi_f, non_umi_f):
        umi2id = {}
        umi_record_iterator, umi_f_type = parse_data(correct_umi_f)
        for item in umi_record_iterator:
            umi = str(item.seq)
            umi2id.setdefault(umi, []).append(str(item.id))

        id2read, ori_file_type = parse_data_dict(non_umi_f)
        raw_records = []
        true_records = []
        for umi in umi2id:
            ids = umi2id[umi]
            # n = 0
            tmp_raw_rec = []
            tmp_true_rec = []
            if len(ids) >= self.config.group_read_number:
                cur_seqs = []
                for id in ids:
                    cur_seqs.append(str(id2read[id].seq))
                seq2count = collections.Counter(cur_seqs)
                max_key = max(seq2count, key=seq2count.get)    
                total_count = sum(list(seq2count.values()))
                max_key_count = seq2count[max_key]
                if max_key_count >= (total_count-max_key_count):     
                    for i in ids:
                        cur_seq = str(id2read[i].seq)
                        dis = editdistance.eval(max_key, cur_seq)
                        desc = "umi:" + str(umi) + "//" + str(id2read[i].description)
                        if dis == 0:
                            # n += 1
                            if ori_file_type == "fastq":
                                rec = SeqRecord(Seq(cur_seq), id=i, description=desc, letter_annotations=id2read[i].letter_annotations)
                            else:
                                rec = SeqRecord(Seq(cur_seq), id=i, description=desc)
                            # raw_records.append(rec)
                            # true_records.append(rec)
                            tmp_raw_rec.append(rec)
                            tmp_true_rec.append(rec)
                        elif dis <= self.config.read_edit_dif:
                        # else:
                            # n += 1
                            if ori_file_type == "fastq":
                                raw_rec = SeqRecord(Seq(cur_seq), id=i, description=desc, letter_annotations=id2read[i].letter_annotations)
                                true_rec = SeqRecord(Seq(max_key), id=i, description=desc, letter_annotations=id2read[i].letter_annotations)
                            else:
                                raw_rec = SeqRecord(Seq(cur_seq), id=i, description=desc)
                                true_rec = SeqRecord(Seq(max_key), id=i, description=desc)
                            # raw_records.append(raw_rec)
                            # true_records.append(true_rec)
                            tmp_raw_rec.append(raw_rec)
                            tmp_true_rec.append(true_rec)
                    # if n >= 10:
                    raw_records.extend(tmp_raw_rec)
                    true_records.extend(tmp_true_rec)                    
        self.logger.info("Write raw and true data to file")
        raw_file = self.config.result_dir + "raw/" + non_umi_f.split('/')[-1]
        true_file = self.config.result_dir + "true/" + non_umi_f.split('/')[-1]
        with open(raw_file, "w") as handle:
            SeqIO.write(raw_records, handle, ori_file_type)        
        with open(true_file, "w") as handle:
            SeqIO.write(true_records, handle, ori_file_type) 
        self.logger.info("Real umi data processing completed.")                 

        return raw_file, true_file

    def raw_true_umi_1(self, correct_umi_f, non_umi_f):
        umi2id = {}
        umi_record_iterator, umi_f_type = parse_data(correct_umi_f)
        umi_n = 0
        for item in umi_record_iterator:
            umi = str(item.seq)
            if 'N' in umi:
                pass
            else:
                umi_n += 1
                umi2id.setdefault(umi, []).append(str(item.id))
        print("test")
        print(umi_n)
        id2read, ori_file_type = parse_data_dict(non_umi_f)
        # print(list(id2read.keys()))
        raw_records = []
        true_records = []
        n = 0
        nn = 0
        for umi in umi2id:
            seq_ids = umi2id[umi]
            tmp_raw_rec = []
            tmp_true_rec = []
            if len(seq_ids) >= 4:
                n += 1
                # print(ids[0])
                # print("#")
                # print(str(id2read[ids[0]].seq))
                cur_seqs = []
                for seq_id in seq_ids:
                    cur_seqs.append(str(id2read[seq_id].seq))
                seq2count = collections.Counter(cur_seqs)
                max_key = max(seq2count, key=seq2count.get)    
                
                total_count = sum(list(seq2count.values()))
                max_key_count = seq2count[max_key]
                print(max_key_count)
                sorted_lst = sorted(list(seq2count.values()), reverse=True)

                # if max_key_count >= (total_count-max_key_count):  
                # if max_key_count >= 0.6 * total_count:  
                if len(sorted_lst) >= 2:
                    if sorted_lst[1] < max_key_count:# and max_key_count >= (total_count-max_key_count):   
                    # if max_key_count >= (total_count-max_key_count):
                    # if max_key_count >= 4:
                        for read_id in seq_ids:
                            cur_seq = str(id2read[read_id].seq)
                            dis = editdistance.eval(max_key, cur_seq)
                            desc = "umi:" + str(umi) + "//" + str(id2read[read_id].description)
                            if dis == 0:
                                nn += 1
                                if ori_file_type == "fastq":
                                    rec = SeqRecord(Seq(cur_seq), id=read_id, description=desc, letter_annotations=id2read[read_id].letter_annotations)
                                else:
                                    rec = SeqRecord(Seq(cur_seq), id=read_id, description=desc)
                                # raw_records.append(rec)
                                # true_records.append(rec)
                                tmp_raw_rec.append(rec)
                                tmp_true_rec.append(rec)
                            # elif dis == 1 or dis == 2:
                            # elif dis <= 10: #and seq2count[cur_seq] < max_key_count
                            else:
                                nn += 1
                                if ori_file_type == "fastq":
                                    raw_rec = SeqRecord(Seq(cur_seq), id=read_id, description=desc, letter_annotations=id2read[read_id].letter_annotations)
                                    true_rec = SeqRecord(Seq(max_key), id=read_id, description=desc, letter_annotations=id2read[read_id].letter_annotations)
                                else:
                                    raw_rec = SeqRecord(Seq(cur_seq), id=read_id, description=desc)
                                    true_rec = SeqRecord(Seq(max_key), id=read_id, description=desc)
                                # raw_records.append(raw_rec)
                                # true_records.append(true_rec)
                                tmp_raw_rec.append(raw_rec)
                                tmp_true_rec.append(true_rec)
                            if dis > 8:
                                print(seq2count.values())
                            # if n >= 10:
                            # raw_records.extend(tmp_raw_rec)
                            # true_records.extend(tmp_true_rec)
                elif len(sorted_lst) == 1:
                    for i in seq_ids:
                        cur_seq = str(id2read[i].seq)
                        dis = editdistance.eval(max_key, cur_seq)
                        desc = "umi:" + str(umi) + "//" + str(id2read[i].description)
                        if dis == 0:
                            n += 1
                            if ori_file_type == "fastq":
                                rec = SeqRecord(Seq(cur_seq), id=i, description=desc, letter_annotations=id2read[i].letter_annotations)
                            else:
                                rec = SeqRecord(Seq(cur_seq), id=i, description=desc)
                            # raw_records.append(rec)
                            # true_records.append(rec)
                            tmp_raw_rec.append(rec)
                            tmp_true_rec.append(rec)
                        # elif dis == 1 or dis == 2:
                        # elif dis <= 4:
                        else:
                            n += 1
                            if ori_file_type == "fastq":
                                raw_rec = SeqRecord(Seq(cur_seq), id=i, description=desc, letter_annotations=id2read[i].letter_annotations)
                                true_rec = SeqRecord(Seq(max_key), id=i, description=desc, letter_annotations=id2read[i].letter_annotations)
                            else:
                                raw_rec = SeqRecord(Seq(cur_seq), id=i, description=desc)
                                true_rec = SeqRecord(Seq(max_key), id=i, description=desc)
                            # raw_records.append(raw_rec)
                            # true_records.append(true_rec)
                            tmp_raw_rec.append(raw_rec)
                            tmp_true_rec.append(true_rec)
                        # if dis > 8:
                        #     print(seq2count.values())
                    # if n >= 10:
            if tmp_raw_rec and tmp_true_rec:
                raw_records.extend(tmp_raw_rec)
                true_records.extend(tmp_true_rec)                                        
        print(n)
        print(nn)
        self.logger.info("Write raw and true data to file")
        raw_file = self.config.result_dir + "raw/" + non_umi_f.split('/')[-1]
        true_file = self.config.result_dir + "true/" + non_umi_f.split('/')[-1]
        with open(raw_file, "w") as handle:
            SeqIO.write(raw_records, handle, ori_file_type)        
        with open(true_file, "w") as handle:
            SeqIO.write(true_records, handle, ori_file_type) 
        self.logger.info("Real umi data processing completed.")                 

        return raw_file, true_file

    def umi_cluster_analysis(self, correct_umi_f, non_umi_f):
        umi2id = {}
        umi_record_iterator, umi_f_type = parse_data(correct_umi_f)
        umi_n = 0
        for item in umi_record_iterator:
            umi = str(item.seq)
            if 'N' in umi:
                pass
            else:
                umi_n += 1
                umi2id.setdefault(umi, []).append(str(item.id))
        print("test")
        print(umi_n)
        id2read, ori_file_type = parse_data_dict(non_umi_f)
        # print(list(id2read.keys()))
        raw_records = []
        true_records = []
        n = 0
        nn = 0
        nnn = 0
        high_seq_umi_number = 0

        umi2high_read2lowread_edit_dis = {}
        umi2high_read2lowread_edit_dis2 = {}
        umi2high_read2lowread_edit_dis3 = {}
        
        read_editDis2count = {}

        for umi in umi2id:
            seq_ids = umi2id[umi]
            # tmp_raw_rec = []
            # tmp_true_rec = []
            high_read2lowread_edit_dis = {}
            high_read2lowread_edit_dis2 = {}
            if len(seq_ids) >= self.config.group_read_number:
                n += 1
                # print(ids[0])
                # print("#")
                # print(str(id2read[ids[0]].seq))
                cur_seqs = []
                for seq_id in seq_ids:
                    cur_seqs.append(str(id2read[seq_id].seq))
                seq2count = collections.Counter(cur_seqs)
                # sorted_lst = sorted(list(seq2count.values()), reverse=True)
                if len(seq2count) >= 2:
                    # seq2count_above = {key: value for key, value in seq2count.items() if value > 4}
                    high_seq2count = {}
                    low_seq2count = {}
                    
                    for seq, count in seq2count.items():
                        if count > self.config.high_freq_thre:
                            high_seq2count[seq] = count
                        else:
                            low_seq2count[seq] = count
                    if len(high_seq2count)==0:
                        high_seq_umi_number += 1
                    # print(high_seq2count)
                    if len(high_seq2count) > 0:
                        
                        high_seqs = list(high_seq2count.keys())

                        for read_id in seq_ids:
                            cur_seq = str(id2read[read_id].seq)
                            desc = "umi:" + str(umi) + "//" + str(id2read[read_id].description)
                            if cur_seq in high_seqs:
                                nn += 1
                                if ori_file_type == "fastq":
                                    rec = SeqRecord(Seq(cur_seq), id=read_id, description=desc, letter_annotations=id2read[read_id].letter_annotations)
                                else:
                                    rec = SeqRecord(Seq(cur_seq), id=read_id, description=desc)
                                raw_records.append(rec)
                                true_records.append(rec)
                                del rec
                                # tmp_raw_rec.append(rec)
                                # tmp_true_rec.append(rec) 
                            else:  
                                sorted_seq2count_above = dict(sorted(high_seq2count.items(), key=lambda item: item[1]), reverse=True)  
                                # print(sorted_seq2count_above)                       
                                for high_read, high_count in sorted_seq2count_above.items():
                                    dis = editdistance.eval(high_read, cur_seq)
                                    if dis <= self.config.read_edit_dif:
                                    # if dis <= 4:
                                        nn += 1
                                        if ori_file_type == "fastq":
                                            raw_rec = SeqRecord(Seq(cur_seq), id=read_id, description=desc, letter_annotations=id2read[read_id].letter_annotations)
                                            true_rec = SeqRecord(Seq(high_read), id=read_id, description=desc, letter_annotations=id2read[read_id].letter_annotations)
                                        else:
                                            raw_rec = SeqRecord(Seq(cur_seq), id=read_id, description=desc)
                                            true_rec = SeqRecord(Seq(high_read), id=read_id, description=desc)
                                        raw_records.append(raw_rec)
                                        true_records.append(true_rec)
                                        del raw_rec, true_rec
                                        # tmp_raw_rec.append(raw_rec)
                                        # tmp_true_rec.append(true_rec) 
                                        break 
                        # for high_read, high_count in high_seq2count.items():
                        for low_seq, low_seq_count in low_seq2count.items():
                            high_seq, edit_dis = self.get_smallest_edit_dis_seq(high_seqs, low_seq)

                            high_read2lowread_edit_dis.setdefault((high_seq, high_seq2count[high_seq]), []).append([low_seq, low_seq_count, edit_dis]) 
                            high_read2lowread_edit_dis2.setdefault((high_seq, high_seq2count[high_seq]), []).append(edit_dis)
                        new_high_read2lowread_edit_dis2 = {}

                        for key, val in high_read2lowread_edit_dis2.items():
                            edit_dis2count = collections.Counter(val)
                            new_high_read2lowread_edit_dis2[key] = edit_dis2count
                            for k, value in edit_dis2count.items():
                                if k in read_editDis2count:
                                    read_editDis2count[k] += value
                                else:
                                    read_editDis2count[k] = value                            
                            
                        new_high_read2lowread_edit_dis = {}
                        for key, val in high_read2lowread_edit_dis.items():
                            temp_dict = {}
                            for item in val:
                                edit_dis = item[2]
                                if edit_dis in temp_dict:
                                    temp_dict[edit_dis] += item[1]
                                else:
                                    temp_dict[edit_dis] = item[1]
                            new_high_read2lowread_edit_dis[key] = temp_dict

                        if len(high_seqs) >= 2:
                            edit_dis_high = self.pairwise_edit_distances(high_seqs)
                            new_high_read2lowread_edit_dis["edit_dis_between_high_freq_read"] = edit_dis_high

                    else:
                        nnn += sum(list(seq2count.values()))
                elif len(seq2count) == 1:
                    for read_id in seq_ids:
                        cur_seq = str(id2read[read_id].seq)
                        desc = "umi:" + str(umi) + "//" + str(id2read[read_id].description)
                        nn += 1
                        if ori_file_type == "fastq":
                            rec = SeqRecord(Seq(cur_seq), id=read_id, description=desc, letter_annotations=id2read[read_id].letter_annotations)
                        else:
                            rec = SeqRecord(Seq(cur_seq), id=read_id, description=desc)
                        # tmp_raw_rec.append(rec)
                        # tmp_true_rec.append(rec)
                        raw_records.append(rec)
                        true_records.append(rec)
                        del rec

            if len(high_read2lowread_edit_dis)>0:
                data_str_keys = {str(key): value for key, value in high_read2lowread_edit_dis.items()}
                
                umi2high_read2lowread_edit_dis[umi] = data_str_keys

                data_str_keys2 = {str(key): value for key, value in new_high_read2lowread_edit_dis2.items()}
                umi2high_read2lowread_edit_dis2[umi] = data_str_keys2

                data_str_keys3 = {str(key): value for key, value in new_high_read2lowread_edit_dis.items()}
                umi2high_read2lowread_edit_dis3[umi] = data_str_keys3



            # if tmp_raw_rec and tmp_true_rec:
            #     raw_records.extend(tmp_raw_rec)
            #     true_records.extend(tmp_true_rec) 
        new_read_editDis2count = {str(key): value for key, value in read_editDis2count.items()}
        with open(self.config.result_dir + non_umi_f.split('/')[-1] + "1.json", "w") as json_file:
            json.dump(umi2high_read2lowread_edit_dis, json_file, indent=4)
        with open(self.config.result_dir + non_umi_f.split('/')[-1] + "2.json", "w") as json_file:
            json.dump(umi2high_read2lowread_edit_dis2, json_file, indent=4)
        with open(self.config.result_dir + non_umi_f.split('/')[-1] + "3.json", "w") as json_file:
            json.dump(umi2high_read2lowread_edit_dis3, json_file, indent=4)
            
        with open(self.config.result_dir + non_umi_f.split('/')[-1] + "edit_dis2number.json", "w") as json_file:
            json.dump(new_read_editDis2count, json_file, indent=4)

        print(n)
        print(nn)
        print(nnn)
        print(high_seq_umi_number)
        self.logger.info("Write raw and true data to file")
        raw_file = self.config.result_dir + "raw/" + non_umi_f.split('/')[-1]
        true_file = self.config.result_dir + "true/" + non_umi_f.split('/')[-1]
        with open(raw_file, "w") as handle:
            SeqIO.write(raw_records, handle, ori_file_type)        
        with open(true_file, "w") as handle:
            SeqIO.write(true_records, handle, ori_file_type) 
        self.logger.info("Real umi data processing completed.")                 

        return raw_file, true_file

    def pairwise_edit_distances(self, strings):
        n = len(strings)
        distances = [[0] * n for _ in range(n)]

        for i in range(n):
            for j in range(i + 1, n):
                distance = editdistance.eval(strings[i], strings[j])
                distances[i][j] = distance
                distances[j][i] = distance

        return distances

    def get_smallest_edit_dis_seq(self, seq_lst, read):
        smallest_edit_dis = 100000
        target_read = ""
        for seq in seq_lst:
            dis = editdistance.eval(seq, read)
            if dis < smallest_edit_dis:
                smallest_edit_dis = dis
                target_read = seq
        return target_read, smallest_edit_dis


    def write_mimic_umis(self, raw_dataset, true_dataset):
        true_records_dict, true_file_type = parse_data_dict(true_dataset)

        true_seqs_lst = []
        # raw_seqs_lst = []
        for name in true_records_dict:
            true_seq = str(true_records_dict[name].seq)
            # raw_seq = str(error_records[name].seq)
            true_seqs_lst.append(true_seq)     
            # raw_seqs_lst.append(raw_seq)
        # raw_read2count = collections.Counter(raw_seqs_lst)
        true_read2count = collections.Counter(true_seqs_lst)
        # write the same read the same id in true dataset, these ids also write to raw dataset
        true_read2id = {}
        i = 1
        for read in true_read2count:
            true_read2id[read] = i
            i += 1
        true_records = []
        raw_records = []
        raw_records_dict, raw_file_type = parse_data_dict(raw_dataset)
        self.logger.info("Read Data")
        for name in tqdm(true_records_dict):
            true_seq = str(true_records_dict[name].seq)
            true_des = "umi:" + str(true_read2id[true_seq]) + "//" + str(true_records_dict[name].description)
            raw_des = "umi:" + str(true_read2id[true_seq]) + "//" + str(raw_records_dict[name].description)
            if true_file_type == "fastq": 
                true_rec = SeqRecord(true_records_dict[name].seq, id=true_records_dict[name].id, description=true_des, letter_annotations=true_records_dict[name].letter_annotations)   
                raw_rec = SeqRecord(raw_records_dict[name].seq, id=raw_records_dict[name].id, description=raw_des, letter_annotations=raw_records_dict[name].letter_annotations)   
            elif true_file_type == "fasta":
                true_rec = SeqRecord(true_records_dict[name].seq, id=true_records_dict[name].id, description=true_des)        
                raw_rec = SeqRecord(raw_records_dict[name].seq, id=raw_records_dict[name].id, description=raw_des)  
            raw_records.append(raw_rec)
            true_records.append(true_rec)
        raw_base_name = raw_dataset.split('/')[-1]
        true_base_name = true_dataset.split('/')[-1]
        umi_raw_dataset = self.config.result_dir + "raw/" + "umi_" + raw_base_name
        umi_true_dataset = self.config.result_dir + "true/" + "umi_"  + true_base_name
        self.logger.info("Rewrite Data")
        with open(umi_raw_dataset, "w") as handle:
            SeqIO.write(raw_records, handle, raw_file_type)        
        with open(umi_true_dataset, "w") as handle:
            SeqIO.write(true_records, handle, true_file_type) 
        self.logger.info("Mimic_umi preprocessing completed.") 
        return umi_raw_dataset, umi_true_dataset
    
    def extract_umis2fasta(self):
        record_iterator, ff_type = parse_data(self.config.input_file) 
        self.logger.info("Read Data")
        umi_records = []
        # non_umi_records = []

        for item in tqdm(record_iterator):
            # seq = str(item.seq)
            seq_id = str(item.id)
            seq_des = str(item.description)
            # print(seq_des)
            umi = seq_des.split(self.config.separator1)[self.config.separator1_idx].split(self.config.separator2)[self.config.separator2_idx]

            umi_rec = SeqRecord(Seq(umi), id=seq_id, description=seq_des)        
            umi_records.append(umi_rec)
                   
        self.logger.info("Rewrite umi and non-umi Data")
        raw_base_name = self.config.input_file.split('/')[-1].split('.')[0]
        umi_raw_dataset = self.config.result_dir + "umi_" + raw_base_name + '.fasta'
        with open(umi_raw_dataset, "w") as handle:
            SeqIO.write(umi_records, handle, 'fasta')        
        self.logger.info("Extract umi to fasta completed.") 
        return umi_raw_dataset, self.config.input_file
    
    def real_umi_in_name_data(self):
        # step1: extract umis to a sperate fasta file
        # umi_raw_dataset, non_umi_raw_dataset = self.extract_umis(original_data, self.config.umi_start, self.config.umi_end, self.config.non_umi_start)
        umi_raw_dataset, non_umi_raw_dataset = self.extract_umis2fasta()
        # step2: correct umi errors
        # self.config.input_file = umi_raw_dataset
        DG = DataGneration(self.logger, self.config)

        genuine_df = DG.extract_umi_genuine_errs(umi_raw_dataset)
        # ##############################################################
        EC = ErrorCorrection(self.logger, self.config)

        corrected_file = EC.umi_correction(umi_raw_dataset, genuine_df)

        raw_file, true_file = self.raw_true_umi(corrected_file, non_umi_raw_dataset)
        
        return raw_file, true_file

    def umi2groundtruth(self):
        raw_file, true_file = self.raw_true_umi(self.config.umi_file, self.config.input_file)
        return raw_file, true_file


# -*- coding: utf-8 -*-
# @Author: Pengyao Ping
# @Date:   2023-02-16 11:04:45
# @Last Modified by:   Pengyao Ping
# @Last Modified time: 2023-09-09 09:58:23

import collections
from Bio import SeqIO
# import gzip
from Bio.SeqRecord import SeqRecord
from Bio.Seq import Seq
from tqdm import tqdm
import copy
import random
from mpire import WorkerPool
import os
from noise2read.utils import *
from noise2read.data_generation import DataGneration
from noise2read.error_orrection import ErrorCorrection
from noise2read.data_preprocessing import DataProcessing
import numpy as np
import sys

class Simulation():
    # def __init__(self, logger, config, num_workers, f_in, error_rate, output_dir, substations, indels, seed=0):
    def __init__(self, logger, config):
        self.logger = logger
        self.config = config
        if os.path.exists(config.result_dir):
            self.logger.info("Directory '% s' already exists" % config.result_dir)
            # for f in os.listdir(self.config.result_dir):
            #     os.remove(os.path.join(self.config.result_dir, f))
        else:
            os.makedirs(config.result_dir)
            self.logger.info("Directory '% s' created" % config.result_dir)

    def enumerate_ed1_seqs(self, read):
        possible_ed1 = []
        if self.config.substations:
            possible_ed1.extend(self.seq2substitution(read))
        # if self.config.indels:
        #     possible_ed1.extend(self.seq2deletion(read))
        #     possible_ed1.extend(self.seq2insertion(read))
        return list(set(possible_ed1))

    def replace_char(self, seq, char, index):
        seq[index] = char
        return ''.join(seq)

    def sub_base(self, base):
        if base == 'A':
            return 'TCG'
        elif base == 'T':
            return 'ACG'
        elif base == 'C':
            return 'ATG'
        elif base == 'G':
            return 'ACT'
        elif base == 'N':
            return 'N'
            
    def seq2substitution(self, read):
        """
        enumerate all the substitutions for one read 
        Args:
            read (str): a sequence
        Returns:
            set: a set contains reads
        """
        editdis1_list = []
        # substitution
        raw_seq = list(read)
        n = len(raw_seq)
        for i in range(n):
            seq = copy.deepcopy(raw_seq)
            temp_base = seq[i]
            #print(temp_base)
            temp_sub_base = list(self.sub_base(temp_base))
            #print(temp_sub_base)
            for b in temp_sub_base:
                if b != 'N':
                    sub_seq = self.replace_char(seq, b, i)
                    editdis1_list.append(sub_seq)
        return set(editdis1_list)

    def seq2deletion(self, read):
        """
        enumerate all the deletions for one read 
        Args:
            read (str): a sequence
        Returns:
            set: a set contains reads
        """
        editdis1_list = []
        seq = list(read)
        n = len(seq)
        # deletion
        for i in range(n):
            del_seq = read[:i] + read[(i+1):]
            editdis1_list.append(del_seq)
        return set(editdis1_list)

    def seq2insertion(self, read):
        """
        enumerate all the insertions for one read 
        Args:
            read (str): a sequence
        Returns:
            set: a set contains reads
        """
        editdis1_list = []
        seq = list(read)
        n = len(seq)
        # insertion
        bases = ['A', 'G', 'C', 'T']
        for i in range(n+1):
            for b in bases:
                raw_seq = copy.deepcopy(seq)
                raw_seq.insert(i, b)
                editdis1_list.append(''.join(raw_seq))
        return set(editdis1_list)
        
    def read2err_read(self, read):
        possible_seqs = self.enumerate_ed1_seqs(read)
        num = len(possible_seqs)
        random.seed(self.config.sim_seed)
        random.shuffle(possible_seqs)
        idx = random.randint(0, num-1)
        return possible_seqs[idx]

    def error_correction(self, config):
        DG = DataGneration(self.logger, config)
        if config.high_ambiguous:
            isolates_file, non_isolates_file, unique_seqs, read_max_len, read_min_len, genuine_df, negative_df, ambiguous_df, high_ambiguous_df = DG.data_files(edit_dis=1)
        else:
            isolates_file, non_isolates_file, unique_seqs, read_max_len, read_min_len, genuine_df, negative_df, ambiguous_df = DG.data_files(edit_dis=1)      
        config.read_max_len = read_max_len
        ###############################################################
        EC = ErrorCorrection(self.logger, config)
        ## one model to predict
        if config.high_ambiguous:
            ##################################
            corrected_file, new_negative_df = EC.all_in_one_correction(isolates_file, non_isolates_file, unique_seqs, genuine_df, negative_df, ambiguous_df, high_ambiguous_df)
        else:
            corrected_file, new_negative_df = EC.all_in_one_correction(isolates_file, non_isolates_file, unique_seqs, genuine_df, negative_df, ambiguous_df, high_ambiguous_df =None) 
        if read_min_len > 30:
            genuine_df2, negative_df2, ambiguous_df2, unique_seqs2 = DG.extract_ed2_errors(corrected_file)
            correct_data = EC.all_in_one_ed2_correction(corrected_file, unique_seqs2, genuine_df2, negative_df2, ambiguous_df2)
        else:
            correct_data = corrected_file
        return correct_data        

    def simplify_error_correction(self, config):
        DG = DataGneration(self.logger, config)
        isolates_file, non_isolates_file, read_max_len, read_min_len, genuine_df, ambiguous_df = DG.simplify_data_files(config.input_file, edit_dis=1)      
        config.read_max_len = read_max_len
        ###############################################################
        EC = ErrorCorrection(self.logger, config)
        corrected_file = EC.simplify_correction(isolates_file, non_isolates_file, genuine_df, ambiguous_df)
            
        if read_min_len > config.min_read_len:
            genuine_df, ambiguous_df = DG.simplify_data_files(corrected_file, edit_dis=2) 
            correct_data = EC.simplify_2nt_correction(corrected_file, genuine_df, ambiguous_df)
        else:
            correct_data = corrected_file
        return correct_data    

    def simulation(self):
        # step 1 correct errors using noise2read
        # 
        # if os.path.exists(self.config.correct_data):
        #     corrected_data = self.config.correct_data
        # else:
            # corrected_data = self.error_correction(self.config)
        corrected_data = self.simplify_error_correction(self.config)
        # inject errors
        raw_data, true_data = self.error_injection(corrected_data)
        # raw_data, true_data = self.error_injection(self.config.input_file)
        DP = DataProcessing( 
            self.logger,
            self.config)
        umi_raw_dataset, umi_true_dataset = DP.write_mimic_umis(raw_data, true_data)
        if os.path.exists(raw_data):
            os.system("rm %s" % raw_data)
        if os.path.exists(true_data):
            os.system("rm %s" % true_data)
        if os.path.exists(corrected_data):
            os.system("rm %s" % corrected_data)
        return umi_raw_dataset, umi_true_dataset
    
    def error_injection(self, data_set):
        records_dict, file_type = parse_data_dict(data_set)
        id_lst = list(records_dict.keys())

        total_seqs_lst = []
        seq2id_dict = {}
        for item in id_lst:
            seq = str(records_dict[item].seq)
            seq2id_dict.setdefault(seq, []).append(item)
            total_seqs_lst.append(seq)
  
        read2counts = collections.Counter(total_seqs_lst)
        select_id_lst = []
        mutated_id_pool = []
        for read, count in read2counts.items():
            if count > self.config.min_freq:
                select_id_lst.extend(seq2id_dict[read])
            if count > self.config.min_read_count:
                mutated_id_pool.extend(seq2id_dict[read])

        if self.config.error_rate1 > 0 and self.config.error_rate1 < 1:
            flag1 = True
            err_1_id_lst = []
            total_id_num = len(mutated_id_pool)
            err_1base_read_num = round(total_id_num * self.config.error_rate1)

            random.shuffle(mutated_id_pool)
            err_1_id_lst.extend(random.sample(mutated_id_pool, err_1base_read_num))
      
            self.logger.info(f"Total candidature reads: {total_id_num}, 1-base-error per read error rate: {self.config.error_rate1}, 1-base-error reads number: {err_1base_read_num}")
        else:
            self.logger.error("Input invalid 1-base-error per read error rate!")
            sys.exit(1)

        # # two base errors
        if self.config.error_rate2 > 0 and self.config.error_rate2 < 1:
            flag2 = True
            remain_id_lst = list(set(mutated_id_pool) - set(err_1_id_lst))
            remain_id_num = len(remain_id_lst)
            err_2bases_read_num = round(remain_id_num * self.config.error_rate2)

            err_2_id_lst = []
            random.shuffle(remain_id_lst)
            err_2_id_lst.extend(random.sample(remain_id_lst, err_2bases_read_num))

            self.logger.info(f"Remaining candidature reads: {remain_id_num}, 2-base-errors per read error rate: {self.config.error_rate2}, 2-base-errors reads number: {err_2bases_read_num}")
        else:
            self.logger.error("Input invalid 2-base-errors per read error rate!")
            sys.exit(1)

        err_records = []
        shared_objects = records_dict, file_type
        if flag1:
            try:
                with WorkerPool(self.config.num_workers, shared_objects=shared_objects, start_method='fork') as pool:
                    for tmp_rec in pool.imap(self.mutate_seq, err_1_id_lst, progress_bar=self.config.verbose):
                        err_records.append(tmp_rec)
            except KeyboardInterrupt:
                # Handle termination signal (Ctrl+C)
                pool.terminate()  # Terminate the WorkerPool before exiting

            except Exception:
                # Handle other exceptions
                pool.terminate()  # Terminate the WorkerPool before exiting
                raise
        if flag2:
            try:
                with WorkerPool(self.config.num_workers, shared_objects=shared_objects, start_method='fork') as pool:
                    for tmp_rec in pool.imap(self.mutate_2base_seq, err_2_id_lst, progress_bar=self.config.verbose):
                        err_records.append(tmp_rec)
            except KeyboardInterrupt:
                # Handle termination signal (Ctrl+C)
                pool.terminate()  # Terminate the WorkerPool before exiting

            except Exception:
                # Handle other exceptions
                pool.terminate()  # Terminate the WorkerPool before exiting
                raise
        del shared_objects
        err_subdataset = self.config.result_dir + "err_subdataset" + file_type
        with open(err_subdataset, "w") as handle:
            SeqIO.write(err_records, handle, file_type)

        ########################################################################################
        err_free_subdataset = self.config.result_dir + "err_free_subdataset" + file_type

        if flag1 and flag2:
            err_id_lst = err_1_id_lst + err_2_id_lst
        elif flag1 and not flag2:
            err_id_lst = err_1_id_lst
        elif not flag1 and flag2:
            err_id_lst = err_2_id_lst
        else:
            self.logger.error("No error id selected, simulation failed!")
            sys.exit(1)            
        err_free_id_lst = list(set(select_id_lst) - set(err_id_lst))
        extract_records(self.config.result_dir, err_free_id_lst, data_set, err_free_subdataset)

        base = self.config.input_file.split("/")[-1]
        file_name = base.split(file_type)[0]
        raw_data = self.config.result_dir + "simulated_raw_" + file_name + file_type
        os.system("cat %s %s > %s" % (err_subdataset, err_free_subdataset, raw_data))
        ##################################################################################
        err_free_subdataset_2 = self.config.result_dir + "err_free_subdataset2" + file_type
        extract_records(self.config.result_dir, err_id_lst, data_set, err_free_subdataset_2)
        true_data = self.config.result_dir + "simulated_true_" + file_name + file_type  
        os.system("cat %s %s > %s" % (err_free_subdataset_2, err_free_subdataset, true_data))
        if os.path.exists(err_free_subdataset_2):
            os.system("rm %s" % err_free_subdataset_2)   
        if os.path.exists(err_free_subdataset):
            os.system("rm %s" % err_free_subdataset)  
        if os.path.exists(err_subdataset):
            os.system("rm %s" % err_subdataset)          
        return raw_data, true_data
    
    def find_first_diff_char(self, string1, string2):
        min_len = min(len(string1), len(string2))

        for i in range(min_len):
            if string1[i] != string2[i]:
                return i

        # If all characters in the shorter string match the corresponding characters in the longer string
        # and the shorter string is a prefix of the longer string, return the length of the shorter string
        if len(string1) != len(string2):
            return min_len

        # If both strings are identical
        return -1

    def mutate_seq(self, shared_objects, idx):
        records_dict, file_type = shared_objects
        ori_seq = str(records_dict[idx].seq)

        possible_seqs = []
        possible_seqs.extend(self.seq2substitution(ori_seq))

        num = len(possible_seqs)
        # random.seed(idx)
        # Do not set seed here because it can produce the same position errors which is not random, or you can set different seed for different sequence
        random.shuffle(possible_seqs)
        ii = random.randint(0, num-1)
        mutation_seq = possible_seqs[ii]

        if file_type == "fastq":
            return SeqRecord(Seq(mutation_seq), id=records_dict[idx].id, description=records_dict[idx].description, letter_annotations=records_dict[idx].letter_annotations)
        elif file_type == "fasta":
            return SeqRecord(Seq(mutation_seq).seq, id=records_dict[idx].id, description=records_dict[idx].description)     

    def mutate_2base_seq(self, shared_objects, idx):
        records_dict, file_type = shared_objects
        ori_seq = str(records_dict[idx].seq)

        possible_seqs = self.enumerate_ed2_seqs(ori_seq)
        num = len(possible_seqs)
        # random.seed(idx)
        # Do not set seed here because it can produce the same position errors which is not random, or you can set different seed for different sequence
        random.shuffle(possible_seqs)
        ii = random.randint(0, num-1)
        mutation_seq = possible_seqs[ii]

        if file_type == "fastq":
            return SeqRecord(Seq(mutation_seq), id=records_dict[idx].id, description=records_dict[idx].description, letter_annotations=records_dict[idx].letter_annotations)
        elif file_type == "fasta":
            return SeqRecord(Seq(mutation_seq).seq, id=records_dict[idx].id, description=records_dict[idx].description)  


    def enumerate_ed2_seqs(self, read):
        possible_ed1 = seq2substitution(read)
        possible_ed2 = []
        for seq in possible_ed1:
            possible_ed2.extend(list(set(seq2substitution(seq)) - possible_ed1))
        return list(set(possible_ed2))

    '''
    def read2_2baseerr_read(self, read):
        possible_seqs = self.enumerate_ed2_seqs(read)
        num = len(possible_seqs)
        random.seed(self.config.sim_seed)
        random.shuffle(possible_seqs)
        idx = random.randint(0, num-1)
        return possible_seqs[idx]
        
    def extract_seq(self, shared_objects, name):
        records_dict, file_type = shared_objects
        if file_type == "fastq":
            return SeqRecord(records_dict[name].seq, id=records_dict[name].id, description=records_dict[name].description, letter_annotations=records_dict[name].letter_annotations)
        elif file_type == "fasta":
            return SeqRecord(records_dict[name].seq, id=records_dict[name].id, description=records_dict[name].description)

    def extract_seqs(self, num_workers, name_lst, data_set, sub_dataset):
        records_dict, file_type = self.parse_data_dict(data_set)
        sub_records = []
        shared_objects = records_dict, file_type
        with WorkerPool(num_workers, shared_objects=shared_objects, start_method='fork') as pool:
            with tqdm(total=len(name_lst), desc=self.logger.info("Extract Seqs")) as pbar:   
                for tmp_rec in pool.imap(self.extract_seq, name_lst):
                    # print(tmp_rec)
                    sub_records.append(tmp_rec)
                    pbar.update() 
        # print(type(sub_records[0]), sub_records[0])
        with open(sub_dataset, "w") as handle:
            SeqIO.write(sub_records, handle, file_type)
        return 
    '''
    '''
    def error_injection(self, data_set):
        records_dict, file_type = parse_data_dict(data_set)
        id_lst = list(records_dict.keys())

        total_seqs_lst = []
        seq2id_dict = {}
        for item in id_lst:
            seq = str(records_dict[item].seq)
            seq2id_dict.setdefault(seq, []).append(item)
            total_seqs_lst.append(seq)
            # total_bases += len(seq)
            # total_reads += 1

        read2counts = collections.Counter(total_seqs_lst)
        select_id_lst = []
        total_bases = 0
        total_reads = 0
        new_read2counts = {}
        for read, count in read2counts.items():
            if count >= 5:
                select_id_lst.extend(seq2id_dict[read])
            if count >= self.config.min_read_count:
                new_read2counts[read] = count
                
                total_reads += count
                total_bases += len(read) * count

        err_reads_num = round(total_bases * self.config.error_rate1)
    
        if total_reads > err_reads_num:
            per_num =  err_reads_num / total_reads
        err_id_lst = []
        for read, count in new_read2counts.items():
            random_num = round(count * per_num)
            name_lst = seq2id_dict[read]
            random.shuffle(name_lst)
            err_id_lst.extend(random.sample(name_lst, random_num))

        # err_reads_num = round(total_bases * self.config.error_rate1 / total_reads)
        self.logger.info(f"total bases: {total_bases}, total reads: {total_reads}, error rate: {self.config.error_rate1}, calculated error reads number: {err_reads_num}, actual error reads number: {len(err_id_lst)}")

        err_subdataset = self.config.result_dir + "err_subdataset" + file_type
        err_records = []
        shared_objects = records_dict, file_type
        with WorkerPool(self.config.num_workers, shared_objects=shared_objects, start_method='fork') as pool:
            for tmp_rec in pool.imap(self.mutate_seq, err_id_lst, progress_bar=self.config.verbose):
                err_records.append(tmp_rec)

        # with WorkerPool(self.config.num_workers, shared_objects=shared_objects, start_method='fork') as pool:
        #     with tqdm(total=len(err_id_lst), desc=self.logger.info("Generate Mutated Seqs")) as pbar:   
        #         for tmp_rec in pool.imap(self.mutate_seq, err_id_lst):
        #             # print(tmp_rec)
        #             err_records.append(tmp_rec)
        #             pbar.update() 
        # print(type(sub_records[0]), sub_records[0])
        with open(err_subdataset, "w") as handle:
            SeqIO.write(err_records, handle, file_type)
        ########################################################################################
        err_free_subdataset = self.config.result_dir + "err_free_subdataset" + file_type

        err_free_id_lst = list(set(select_id_lst) - set(err_id_lst))
        extract_records(self.config.result_dir, err_free_id_lst, data_set, err_free_subdataset)

        base = self.config.input_file.split("/")[-1]
        file_name = base.split(file_type)[0]
        raw_data = self.config.result_dir + "simulated_raw_" + file_name + file_type
        os.system("cat %s %s > %s" % (err_subdataset, err_free_subdataset, raw_data))
        ##################################################################################
        err_free_subdataset_2 = self.config.result_dir + "err_free_subdataset2" + file_type
        extract_records(self.config.result_dir, err_id_lst, data_set, err_free_subdataset_2)
        true_data = self.config.result_dir + "simulated_true_" + file_name + file_type  
        os.system("cat %s %s > %s" % (err_free_subdataset_2, err_free_subdataset, true_data))     
        return raw_data, true_data
    '''

    '''
    def mutate_seq(self, shared_objects, idx):
        records_dict, file_type = shared_objects
        ori_seq = str(records_dict[idx].seq)
        mutation_seq = self.read2err_read(ori_seq)
        if file_type == "fastq":
            # dis = editdistance.eval(ori_seq, mutation_seq)
            # if dis == 0:
            #     pass
            # elif dis == 1:
            #     indel_idx = self.find_first_diff_char(ori_seq, mutation_seq)
            #     # Get the quality scores list from letter_annotations
            #     quality_scores = records_dict[idx].letter_annotations["phred_quality"]
            #     if len(ori_seq) < len(mutation_seq):
            #         # Calculate the average score
            #         min_score = min(quality_scores)
            #         # Insert the average score at the specified index
            #         quality_scores.insert(indel_idx, min_score)
            #         records_dict[idx].letter_annotations["phred_quality"] = quality_scores
            #     elif len(ori_seq) > len(mutation_seq):
            #         quality_scores.pop(indel_idx)
            #         records_dict[idx].letter_annotations["phred_quality"] = quality_scores
            # else:
            #     self.logger.error("Generate non 1-base error read.")
            #     sys.exit(1)
            return SeqRecord(Seq(mutation_seq), id=records_dict[idx].id, description=records_dict[idx].description, letter_annotations=records_dict[idx].letter_annotations)
        elif file_type == "fasta":
            return SeqRecord(Seq(mutation_seq).seq, id=records_dict[idx].id, description=records_dict[idx].description) 
    '''

    ''' 
    def error_injection(self, data_set):
        records_dict, file_type = parse_data_dict(data_set)
        id_lst = list(records_dict.keys())
        total_seqs_lst = []
        seq2id_dict = {}
        for item in id_lst:
            seq = str(records_dict[item].seq)
            seq2id_dict.setdefault(seq, []).append(item)
            total_seqs_lst.append(seq)
            # total_bases += len(seq)
            # total_reads += 1
            
        read2counts = collections.Counter(total_seqs_lst)
        select_id_lst = []
        # total_bases = 0
        total_reads = 0
        new_read2counts = {}
        for read, count in read2counts.items():
            # if count >= 5:
            if count > self.config.min_freq:
                select_id_lst.extend(seq2id_dict[read])
            if count > self.config.min_read_count:
                new_read2counts[read] = count
                total_reads += count
                # total_bases += len(read) * count
                
        # one base error
        err_1_base_num = round(total_reads * self.config.error_rate1)
        if total_reads > err_1_base_num:
            per_num =  err_1_base_num / total_reads
            err_1_id_lst = []
            for read, count in new_read2counts.items():
                random_num = round(count * per_num)
                name_lst = seq2id_dict[read]

                rng = np.random.default_rng(seed=self.config.sim_seed)
                indices = rng.permutation(len(name_lst))[:random_num]
                # print(indices)
                err_1_id_lst.extend(name_lst[i] for i in indices)
                # print(err_1_id_lst)
        else:
            self.logger.error("1 base error reads number is larger than the total reads number!")
            sys.exit(1)

        # two base errors
        if self.config.error_rate2 > 0:
            err_2_id_lst = []
            cur_total_reads = total_reads - len(err_1_id_lst)
            err_2_base_num = round(cur_total_reads * self.config.error_rate2)

            if cur_total_reads > err_2_base_num:
                per_num_2base =  err_2_base_num / (cur_total_reads * 2)
                for read, count in new_read2counts.items():
                    random_num_2base = round(count * per_num_2base)
                    cur_name_lst = list(set(seq2id_dict[read]) - set(err_1_id_lst))

                    rng = np.random.default_rng(seed=self.config.sim_seed)
                    indices = rng.permutation(len(cur_name_lst))[:random_num_2base]
                    err_2_id_lst.extend(cur_name_lst[i] for i in indices)
                    
                self.logger.info(f"Total reads: {total_reads}, 1 base error per read error rate: {self.config.error_rate1}, 1 base error reads number: {len(err_1_id_lst)}, 2 bases per read error rate: {self.config.error_rate2}, 2 bases error reads number: {len(err_2_id_lst)}")
            else:
                self.logger.error("2 bases error reads number is larger than the total reads number!")
                sys.exit(1)
        else:
            # err_reads_num = round(total_bases * self.config.error_rate / total_reads)
            self.logger.info(f"Total reads: {total_reads}, 1 base error per read error rate: {self.config.error_rate1}, 1 base error reads number: {len(err_1_id_lst)}")
        
        err_1base_records = []
        shared_objects = records_dict, file_type
        with WorkerPool(self.config.num_workers, shared_objects=shared_objects, start_method='fork') as pool:
            for tmp_rec in pool.imap(self.mutate_seq, err_1_id_lst, progress_bar=self.config.verbose):
                err_1base_records.append(tmp_rec)

        err_2base_records = []
        shared_objects = records_dict, file_type
        with WorkerPool(self.config.num_workers, shared_objects=shared_objects, start_method='fork') as pool:
            for tmp_rec in pool.imap(self.mutate_2base_seq, err_2_id_lst, progress_bar=self.config.verbose):
                err_2base_records.append(tmp_rec)

        err_records = err_1base_records + err_2base_records
        err_subdataset = self.config.result_dir + "err_subdataset" + file_type
        with open(err_subdataset, "w") as handle:
            SeqIO.write(err_records, handle, file_type)

        ########################################################################################
        err_free_subdataset = self.config.result_dir + "err_free_subdataset" + file_type

        err_id_lst = err_1_id_lst + err_2_id_lst
        err_free_id_lst = list(set(select_id_lst) - set(err_id_lst))
        extract_records(self.config.result_dir, err_free_id_lst, data_set, err_free_subdataset)

        base = self.config.input_file.split("/")[-1]
        file_name = base.split(file_type)[0]
        raw_data = self.config.result_dir + "simulated_raw_" + file_name + file_type
        os.system("cat %s %s > %s" % (err_subdataset, err_free_subdataset, raw_data))
        ##################################################################################
        err_free_subdataset_2 = self.config.result_dir + "err_free_subdataset2" + file_type
        extract_records(self.config.result_dir, err_id_lst, data_set, err_free_subdataset_2)
        true_data = self.config.result_dir + "simulated_true_" + file_name + file_type  
        os.system("cat %s %s > %s" % (err_free_subdataset_2, err_free_subdataset, true_data))     
        return raw_data, true_data
    '''

    '''
    total_seqs_lst = []
    seq2id_dict = {}
    for item in id_lst:
        seq = str(records_dict[item].seq)
        seq2id_dict.setdefault(seq, []).append(item)
        total_seqs_lst.append(seq)
        # total_bases += len(seq)
        # total_reads += 1

    read2counts = collections.Counter(total_seqs_lst)
    select_id_lst = []
    total_bases = 0
    total_reads = 0
    new_read2counts = {}
    for read, count in read2counts.items():
        if count >= 5:
            select_id_lst.extend(seq2id_dict[read])
        if count >= self.config.min_read_count:
            new_read2counts[read] = count
            
            total_reads += count
            total_bases += len(read) * count

    err_reads_num = round(total_bases * self.config.error_rate1)

    if total_reads > err_reads_num:
        per_num =  err_reads_num / total_reads
    err_1_id_lst = []
    for read, count in new_read2counts.items():
        random_num = round(count * per_num)
        name_lst = seq2id_dict[read]
        random.shuffle(name_lst)
        err_1_id_lst.extend(random.sample(name_lst, random_num))

    # err_reads_num = round(total_bases * self.config.error_rate1 / total_reads)
    self.logger.info(f"total bases: {total_bases}, total reads: {total_reads}, error rate: {self.config.error_rate1}, calculated error reads number: {err_reads_num}, actual error reads number: {len(err_1_id_lst)}")
    '''


# if __name__ == '__main__':
#     logger = custom_logger("simulation", debug_mode=True)
#     input_file = "/home/pping/Data/Repo/data/noise2read_data/group3/SAS-Cov-2/noise2read_result/new_no_ml/output/umi_simulated_raw_original_sars_r1_corrected_correct.fastq"
#     sim = Simulation(logger, 30, )

#     raw_data, true_data = sim.error_injection(input_file)
#     output_dir = "/home/pping/Data/Repo/data/noise2read_data/group3/SAS-Cov-2/"
#     DP = DataProcessing( 
#         logger,
#         num_workers=30,
#         output_dir=output_dir)
#     umi_raw_dataset, umi_true_dataset = DP.write_mimic_umis(raw_data, true_data)


'''
@File    :   umitest.py
@Time    :   2022/10/25 23:21:30
@Author  :   Pengyao Ping
@Version :   1.0
@Contact :   Ping.Pengyao@gmail.com
@desc    : 
'''

# from collections import Counter
import collections
import math
from Bio import SeqIO
# import os
# import editdistance
# import xlsxwriter
# from noise2read.utils import parse_file_type

def parse_file_type(data_set):
    items = data_set.split(".")
    ext = items[-1]
    if ext == 'fa' or ext == 'fasta':
        f_type = 'fasta'
    elif ext == 'fq' or ext == 'fastq':
        f_type = 'fastq'
    elif ext == 'gz':
        f_type = items[-2] + '.' + ext
    return f_type

# raw_dataset = '/home/pping/Data/Repo/data/noise2read_data/umi-data/group1/umi24/raw/non_umi_SRR1543964.fastq'
# ground_truth_dataset = '/home/pping/Data/Repo/data/noise2read_data/umi-data/group1/umi24/true/non_umi_SRR1543964.fastq'
def data_analysis(raw_dataset, ground_truth_dataset):
    true_records = SeqIO.index(ground_truth_dataset, parse_file_type(ground_truth_dataset))
    error_records = SeqIO.index(raw_dataset, parse_file_type(raw_dataset))
    # correct_records = SeqIO.index(correct_dataset, parse_file_type(correct_dataset))
    total_reads_num = 0
    # for calculating entropy
    correct_errfree_seqs = []
    correct_err_seqs = []
    raw_errfreee_seqs = []
    raw_err_seqs = []

    true_seqs_lst = []
    raw_seqs_lst = []
    # correct_seqs_lst = []
    true_umi2seqs = {}
    raw_umi2seqs = {}
    # correct_umi2seqs = {}

    true_seq2umi = {}
    for name in error_records:
        true_seq = str(true_records[name].seq)
        raw_seq = str(error_records[name].seq)
        # correct_seq = str(correct_records[name].seq) 

        umi_base = str(true_records[name].description).split('//')[0]
        umi = umi_base.split(':')[1]
        true_umi2seqs.setdefault(umi, []).append(true_seq)
        raw_umi2seqs.setdefault(umi, []).append(raw_seq)
        # correct_umi2seqs.setdefault(umi, []).append(correct_seq)

        true_seq2umi.setdefault(true_seq, []).append(umi)

        true_seqs_lst.append(true_seq)     
        raw_seqs_lst.append(raw_seq)
        # correct_seqs_lst.append(correct_seq)

    raw_read2count = collections.Counter(raw_seqs_lst)
    true_read2count = collections.Counter(true_seqs_lst)
    # correct_read2count = collections.Counter(correct_seqs_lst)

    num = 0
    # read_num = 0
    for seq in true_seq2umi:
        unique_umi = set(true_seq2umi[seq])
        if len(unique_umi) > 1:
            num += 1
            print(unique_umi)
    print(num)

if __name__ == '__main__':

    # raw_dataset = '/home/pping/Data/Repo/data/noise2read_data/umi-data/group1/umi24/raw/non_umi_SRR1543964.fastq'
    # ground_truth_dataset = '/home/pping/Data/Repo/data/noise2read_data/umi-data/group1/umi24/true/non_umi_SRR1543964.fastq'

    raw_dataset = '/home/pping/Data/Repo/data/noise2read_data/umi-data/group1/test12/raw/non_umi_SRR1543964.fastq'
    ground_truth_dataset = '/home/pping/Data/Repo/data/noise2read_data/umi-data/group1/test12/true/non_umi_SRR1543964.fastq'

    # data_analysis(raw_dataset, ground_truth_dataset)

    umi_data = '/home/pping/Data/Repo/data/noise2read_data/umi-data/group1/umi1224/umi_SRR1543964_corrected.fastq'

    umi_records = SeqIO.index(umi_data, parse_file_type(umi_data))
    umi_seqs = []
    for name in umi_records:
        umi_seqs.append(str(umi_records[name].seq))
    
    read2counts = collections.Counter(umi_seqs)
    print(read2counts)


# -*- coding: utf-8 -*-
# @Author: Pengyao Ping
# @Date:   2023-02-16 11:01:06
# @Last Modified by:   Pengyao Ping
# @Last Modified time: 2023-04-07 17:22:18

from Bio import SeqIO
import gzip
# from Bio.SeqRecord import SeqRecord
# from Bio.Seq import Seq
import logging
from colorlog import ColoredFormatter
import copy
import random
import editdistance
import os
import datetime
import tracemalloc

class MemoryMonitor:
    def __init__(self, logger):
        self.current_memory = 0
        self.peak_memory = 0
        self.traceback = None
        self.logger = logger

    def start(self):
        tracemalloc.start()

    def stop(self):
        tracemalloc.stop()

    def measure(self):
        current, peak = tracemalloc.get_traced_memory()
        self.current_memory = current
        if peak > self.peak_memory:
            self.peak_memory = peak

        self.logger.info(f"Current memory usage: {self.current_memory / (1024 * 1024)} MB")
        self.logger.info(f"Recent peak memory usage: {self.peak_memory / (1024 * 1024)} MB")

    def get_memory_traceback(self, obj):
        self.traceback = tracemalloc.get_object_traceback(obj)
        self.logger.debug("Memory Allocation Traceback:")
        for filename, lineno, funcname, line in self.traceback:
            self.logger.debug(f"  File '{filename}', line {lineno}, in {funcname}")
            if line:
                self.logger.debug(f"    {line.strip()}")


def custom_logger(root_name, debug_mode) -> logging.Logger: 
    logger = logging.getLogger(root_name)    
    
    formatter = ColoredFormatter(
        "%(green)s[%(asctime)s] %(blue)s%(name)s %(log_color)s%(levelname)-8s%(reset)s %(message)s",
        datefmt=None,
        reset=True,
        log_colors={
            'DEBUG':    'cyan',
            'INFO':     'green',
            'WARNING':  'yellow',
            'ERROR':    'red',
            'CRITICAL': 'red, bg_white',
        },
        secondary_log_colors={},
        style='%'
    )
    if debug_mode:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)
    stream_handler = logging.StreamHandler()
    stream_handler.setFormatter(formatter)
    logger.addHandler(stream_handler)
    
    # Output full log
    file_handler = logging.FileHandler( datetime.datetime.now().strftime("%Y_%m_%d_%H_%M_") + 'noise2read.log')

    file_handler.setLevel(logging.INFO)

    # formatter = logging.Formatter(log_format)
    logger.addHandler(file_handler)

    # # Output warning log
    # file_handler = logging.FileHandler('noise2read.Warning.log')
    # file_handler.setLevel(logging.WARNING)
    # # formatter = logging.Formatter(log_format)
    # file_handler.setFormatter(formatter)
    # logger.addHandler(file_handler)

    # # Output error log
    # file_handler = logging.FileHandler('noise2read.Error.log')
    # file_handler.setLevel(logging.ERROR)
    # # formatter = logging.Formatter(log_format)
    # file_handler.setFormatter(formatter)
    # logger.addHandler(file_handler)

    return logger

def parse_file_type(data_set):
    items = data_set.split(".")
    ext = items[-1]
    if ext == 'fa' or ext == 'fasta':
        f_type = 'fasta'
    elif ext == 'fq' or ext == 'fastq':
        f_type = 'fastq'
    elif ext == 'gz':
        f_type = items[-2] + '.' + ext
    return f_type

def parse_data(data_set):
    file_type = parse_file_type(data_set)
    if file_type == 'fastq.gz' or file_type == 'fq.gz' or file_type == 'fa.gz' or file_type == 'fasta.gz':
        ff_type = file_type.split('.')[0]
        handle = gzip.open(data_set, 'rt')
        record_iterator = SeqIO.parse(handle, ff_type)
        return record_iterator, ff_type
    else:
        record_iterator = SeqIO.parse(data_set, file_type) 
        return record_iterator, file_type

def parse_data_index(data_set):
    file_type = parse_file_type(data_set)
    if file_type == 'fastq.gz' or file_type == 'fq.gz' or file_type == 'fa.gz' or file_type == 'fasta.gz':
        ff_type = file_type.split('.')[0]
        handle = gzip.open(data_set, 'rt')
        records = SeqIO.index(handle, ff_type)
        return records, ff_type
    else:
        records = SeqIO.index(data_set, file_type)     
        return records, file_type

def parse_data_dict(data_set):
    file_type = parse_file_type(data_set)
    if file_type == 'fastq.gz' or file_type == 'fq.gz' or file_type == 'fa.gz' or file_type == 'fasta.gz':
        ff_type = file_type.split('.')[0]
        handle = gzip.open(data_set, 'rt')
        records_dict = SeqIO.to_dict(SeqIO.parse(handle, ff_type))
        return records_dict, ff_type
    else:
        records_dict = SeqIO.to_dict(SeqIO.parse(data_set, file_type))   
        return records_dict, file_type
'''
def extract_seq(shared_objects, name):
    records_dict, file_type = shared_objects
    if file_type == "fastq":
        return SeqRecord(records_dict[name].seq, id=records_dict[name].id, description=records_dict[name].description, letter_annotations=records_dict[name].letter_annotations)
    elif file_type == "fasta":
        return SeqRecord(records_dict[name].seq, id=records_dict[name].id, description=records_dict[name].description)

def extract_seqs(num_workers, name_lst, data_set, sub_dataset, logger):
    records_dict, file_type = parse_data_dict(data_set)
    sub_records = []
    shared_objects = records_dict, file_type
    with WorkerPool(num_workers, shared_objects=shared_objects, start_method='fork') as pool:
        with tqdm(total=len(name_lst), desc=logger.info("Extract Seqs")) as pbar:   
            for tmp_rec in pool.imap(extract_seq, name_lst):
                # print(tmp_rec)
                sub_records.append(tmp_rec)
                pbar.update() 
    # print(type(sub_records[0]), sub_records[0])
    with open(sub_dataset, "w") as handle:
        SeqIO.write(sub_records, handle, file_type)
    return 
'''
def extract_records(working_dir, name_lst, data_set, sub_dataset):
    f_name_lst = os.path.join(working_dir, 'tmp_name.lst')
    with open(f_name_lst, "w") as outfile:
        outfile.write("\n".join(name_lst))
    os.system("seqtk subseq %s %s > %s" % (data_set, f_name_lst, sub_dataset))
    os.system("rm %s" % f_name_lst)
    # record_iterator, file_type = parse_data(data_set)
    # records = [rec for rec in record_iterator if str(rec.id) in name_lst]
    # SeqIO.write(records, sub_dataset, file_type)
    return
#####################################################################################
def sub_base(base):
    if base == 'A':
        return 'TCG'
    elif base == 'T':
        return 'ACG'
    elif base == 'C':
        return 'ATG'
    elif base == 'G':
        return 'ACT'
    elif base == 'N':
        return 'N'

def replace_char(seq, char, index):
    seq[index] = char
    return ''.join(seq)

def seq2substitution(read):
    """
    enumerate all the substitutions for one read 
    Args:
        read (str): a sequence
    Returns:
        set: a set contains reads
    """
    editdis1_list = []
    # substitution
    raw_seq = list(read)
    n = len(raw_seq)
    for i in range(n):
        seq = copy.deepcopy(raw_seq)
        temp_base = seq[i]
        #print(temp_base)
        temp_sub_base = list(sub_base(temp_base))
        #print(temp_sub_base)
        for b in temp_sub_base:
            if b != 'N':
                sub_seq = replace_char(seq, b, i)
                editdis1_list.append(sub_seq)
    return set(editdis1_list)

def seq2deletion(read):
    """
    enumerate all the deletions for one read 
    Args:
        read (str): a sequence
    Returns:
        set: a set contains reads
    """
    editdis1_list = []
    seq = list(read)
    n = len(seq)
    # deletion
    for i in range(n):
        del_seq = read[:i] + read[(i+1):]
        editdis1_list.append(del_seq)
    return set(editdis1_list)

def seq2insertion(read):
    """
    enumerate all the insertions for one read 
    Args:
        read (str): a sequence
    Returns:
        set: a set contains reads
    """
    editdis1_list = []
    seq = list(read)
    n = len(seq)
    # insertion
    bases = ['A', 'G', 'C', 'T']
    for i in range(n+1):
        for b in bases:
            raw_seq = copy.deepcopy(seq)
            raw_seq.insert(i, b)
            editdis1_list.append(''.join(raw_seq))
    return set(editdis1_list)

def enumerate_ed1_seqs(read):
    possible_ed1 = []
    possible_ed1.extend(seq2deletion(read))
    possible_ed1.extend(seq2substitution(read))
    possible_ed1.extend(seq2insertion(read))
    return set(possible_ed1)

def enumerate_ed2_seqs(read):
    # possible_ed1 = self.enumerate_ed1_seqs(read)
    possible_ed1 = seq2substitution(read)
    possible_ed2 = []
    for seq in possible_ed1:
        possible_ed2.extend(list(set(seq2substitution(seq)) - possible_ed1))
    return set(possible_ed2)
    
def random_ed2_seq(read, total_reads, num):
    '''
    return num reads that each have two bases difference from given read and these generated reads not exist in total_reads
    '''
    editdis2_list = []
    # substitution
    raw_seq = list(read)
    n = len(raw_seq)
    pos_lst = []
    for i in range(num):
        pos_lst.append(random.sample(range(n), 2))
    for item in pos_lst:
        seq = copy.deepcopy(raw_seq)
        i = item[0]
        j = item[1]
        i_base = seq[i]
        j_base = seq[j]
        i_sub_base = random.sample(list(sub_base(i_base)), 1)[0]    
        j_sub_base = random.sample(list(sub_base(j_base)), 1)[0]  
        sub_seq = replace_char(seq, i_sub_base, i)
        sub_seq = replace_char(seq, j_sub_base, j)
        if sub_seq not in total_reads:
            editdis2_list.append(sub_seq)
    return editdis2_list

def error_type_classification(read1, read2):
    f_len = len(read1)
    s_len = len(read2)  
    # position = -1  
    dis = editdistance.eval(read1, read2)
    # print(dis)
    if dis == 1:              
        if f_len == s_len:
            position = -1
            for index in range(f_len):
                if read1[index] == read2[index]:
                    continue
                else:
                    position = index
                    # first = f_seq[index]
                    # second = s_seq[index]
                    break
            first = read1[position]
            second = read2[position]
            if position == 0:
                f_kmer = read1[0:2]
                s_kmer = read2[0:2]
            elif position == f_len:
                f_kmer = read1[-2:] 
                s_kmer = read2[-2:]
            else:
                f_kmer = read1[position-1 : position+2]
                s_kmer = read2[position-1 : position+2]  

        elif f_len < s_len:
            position = -1
            num = 0
            for index in range(f_len):
                if read1[index] == read2[index]:
                    num = num + 1
                else:
                    position = index
                    break
            if num == f_len:
                position = f_len
            first = 'X'
            second = read2[position]  
            if position == 0:
                f_kmer = 'X' + read1[0]
                s_kmer = read2[0:2]
            elif position == f_len:
                f_kmer = read1[-1] + 'X'
                s_kmer = read2[-2:]
            else:
                f_kmer = read1[position-1] + 'X' + read1[position]
                s_kmer = read2[position-1 : position+2]                    
        elif f_len > s_len:
            position = -1
            num = 0
            for index in range(s_len):
                if read1[index] == read2[index]:
                    num = num + 1
                else:
                    position = index
                    break
            if num == s_len:
                position = s_len
            first = read1[position]
            second = 'X'
            if position == 0:
                f_kmer = read1[0:2]
                s_kmer = 'X' + read2[0]
            elif position == s_len:
                f_kmer = read1[-2:] 
                s_kmer = read2[-1] + 'X'
            else:
                f_kmer = read1[position-1 : position+2]
                s_kmer = read2[position-1] + 'X' + read2[position]    

        errorType = first + '-' + second
    
        return [errorType, f_kmer, s_kmer]
    else:
        raise ValueError("The editdistance of two reads in the input list must equal to one!")

def usage():
    # print(" ")
    print("noise2read Usage:")
    print("   Mandatory:")
    print("     -m|--module                   module selection")
    print("   Modules: [correction, amplicon_correction, umi_correction, mimic_umi, real_umi, evaluation, simulation]")
    # print(" ")
    print("1. Using config file")
    print("     noise2read -m|--module <module_name> -c <path_configuration_file>")
    print("   Mandatory:")
    print("     -c|--config                   input configuration file")
    # print(" ")
    print("2. Using command line with the default parameters")
    print("     noise2read -m|--module <module_name> -i <path_raw_data.fastq|fasta|fa|fq>")
    print("   Mandatory:")
    print("     -i|--input                    input raw data to be corrected")
    print("   Options:")
    print("     -d|--directory                set output directory")
    print("     -a|--high_ambiguous           predict high ambiguous errors using machine learning when set true, defaut true")
    print("     -t|--true                     input ground truth data if you have")
    print("     -r|--rectification            input corrected data when using module evaluation")
    print("     -p|--parallel                 use multiple cpu cores, default total cpu cores - 2")
    print("     -g|--tree_method              use gpu for training and prediction, default auto, (options gpu_hist, hist, auto)")
    # print("     -o|--over_sampling            use over sampling or downsampling for negative samples, default True")
    print("     -h|--help                     show this help")
    # print("########################################################################################################################")

#####################################################################################
# def split_seqs(num_workers, name_set, data_set, sub_dataset_1, sub_dataset_2):
#     records_dict, file_type = parse_data_dict(data_set)
#     sub_records_1 = []
#     sub_records_2 = []
#     shared_objects = records_dict, file_type

#     with WorkerPool(num_workers, shared_objects=shared_objects, start_method='fork') as pool:
#         with tqdm(total=len(name_lst), desc="Extract Seqs") as pbar:   
#             for tmp_rec in pool.imap(extract_seq, name_lst):
#                 # print(tmp_rec)
#                 sub_records.append(tmp_rec)
#                 pbar.update() 

#     for item in record_iterator:
#         qual = {}
#         name = item.id
#         if file_type == "fastq":
#             qual['phred_quality'] = item.letter_annotations['phred_quality']
#             tmp_rec = SeqRecord(item.seq, id=name, description=item.description, letter_annotations=qual) 
#         elif file_type == "fasta":
#             tmp_rec = SeqRecord(item.seq, id=name, description=item.description)    
#         if name in name_set:             
#             sub_records_1.append(tmp_rec)
#         elif name in non_name_set:                        
#             sub_records_2.append(tmp_rec)
#     with open(sub_dataset_1, "w") as handle:
#         SeqIO.write(sub_records_1, handle, file_type)
#     with open(sub_dataset_2, "w") as handle:
#         SeqIO.write(sub_records_2, handle, file_type)
#     return 

if __name__ == "__main__":
    # input_file = '/home/pping/Data/Repo/data/noise2read_data/group1/raw/new_tcr.seq.real_SRR1543964.fastq'
    
    # records_dict, file_type = parse_data_dict(input_file)
    # sub_records = []
    # for name in records_dict:
    #     if file_type == "fastq":
    #         # qual = {}
    #         # qual['phred_quality'] = records[name].letter_annotations["phred_quality"]
    #         tmp_rec = SeqRecord(records_dict[name].seq, id=records_dict[name].id, description=records_dict[name].description, letter_annotations=records_dict[name].letter_annotations) 
    #         print(type(records_dict[name].seq))
    #         print(type(records_dict[name].id))
    #         print(type(records_dict[name].description))
    #         print(type(records_dict[name].letter_annotations))
    #     elif file_type == "fasta":
    #         tmp_rec = SeqRecord(records_dict[name].seq, id=records_dict[name].id, description=records_dict[name].description)      
    #         print(records_dict[name].seq, records_dict[name].id)
    #     break
    num = 3
    total_reads = ['AACGT', 'GACGT', 'CACGT', 'TACGT', 'AACGT', 'AGCGT','ACCGT', 'ATCGT', 'ACAGT', 'ACGGT', 'ACCGT', 'ACTGT', 'ACGAT']
    read = 'AACGT'
    seqs = random_ed2_seq(read, total_reads, num)
    print(seqs)


# -*- coding: utf-8 -*-
# @Author: Pengyao Ping
# @Date:   2022-12-29 23:04:12
# @Last Modified by:   Pengyao Ping
# @Last Modified time: 2023-09-07 16:52:11

from noise2read.config import Config
import sys, getopt
from noise2read.data_generation import DataGneration
from noise2read.error_orrection import ErrorCorrection
from noise2read.data_analysis import DataAnalysis
import os
from noise2read.simulation import Simulation
from noise2read.utils import custom_logger, usage
from noise2read.data_preprocessing import DataProcessing
# from contextlib import redirect_stdout
import noise2read
from noise2read.utils import MemoryMonitor

def main():
    argv = sys.argv[1:]
    # create logger
    logger = custom_logger("noise2read", debug_mode=True)
    MM = MemoryMonitor(logger)
    MM.start()
    ##############################################################
    try:
        # opts, args = getopt.getopt(argv, "m:c:i:u:t:r:d:p:a:g:o:l:h:", ["module=", "config=", "input=", "umi_file", "true", "rectification", "directory", "parallel", "high_ambiguous", "tree_method", "over_sampling", "libray_layout", "help"]) 
        # opts, args = getopt.getopt(argv, "m:c:i:u:t:r:d:p:a:g:l:hv", ["module=", "config=", "input=", "umi_file", "true", "rectification", "directory", "parallel", "high_ambiguous", "tree_method", "libray_layout", "help", "version"]) 
        opts, args = getopt.getopt(argv, "m:c:i:u:t:r:d:p:a:g:hv", ["module=", "config=", "input=", "umi_file", "true", "rectification", "directory", "parallel", "high_ambiguous", "tree_method", "help", "version", "tau="]) 
        script_name = sys.argv[0]
        input_commands = [script_name]

        for opt, arg in opts:
            input_commands.append(opt)
            if arg:
                input_commands.append(arg)

        for arg in args:
            input_commands.append(arg)
        logger.info("Commands: " + " ".join(input_commands))
        
        if opts:
            opts_dict = dict(opts)
            opts_keys = list(opts_dict.keys())
            tar_set = set(opts_keys)

            h_lst = list({'-h', '--help'}.intersection(tar_set))
            v_lst = list({'-v', '--version'}.intersection(tar_set))
            if h_lst:
                usage()           
                sys.exit() 
            elif v_lst: 
                print("Version: " + noise2read.__version__)
                sys.exit()
            else: 
                m_lst = list({"-m", "--module"}.intersection(tar_set))
                c_lst = list({"-c", "--config"}.intersection(tar_set)) 
                i_lst = list({"-i", "--input"}.intersection(tar_set)) 
                u_lst = list({"-u", "--umi_file"}.intersection(tar_set)) 
                t_lst = list({"-t", "--true"}.intersection(tar_set)) 
                r_lst = list({"-r", "--rectification"}.intersection(tar_set))   
                d_lst = list({"-d", "--directory"}.intersection(tar_set))
                p_lst = list({"-p", "--parallel"}.intersection(tar_set))
                a_lst = list({"-a", "--high_ambiguous"}.intersection(tar_set))
                g_lst = list({"-g", "--tree_method"}.intersection(tar_set))
                tau_lst = list({"--tau"}.intersection(tar_set))
                # print(tau_lst)
                # o_lst = list({"-o", "--over_sampling"}.intersection(tar_set))

                # l_lst = list({"-l", "--libray_layout"}.intersection(tar_set))
                # ref_lst = list({"-f", "--reference_in"}.intersection(tar_set))
                # r0_lst = list({"-0", "--read"}.intersection(tar_set))
                # r1_lst = list({"-1", "--read1"}.intersection(tar_set))
                # r2_lst = list({"-2", "--read1"}.intersection(tar_set))
                # align_lst = list({"-A", "--Alignment"}.intersection(tar_set))
                    
                if m_lst:
                    module_arg = opts_dict[m_lst[0]]
                        
                else:
                    logger.error("Not select any module, using command noise2read -h/--help for usage.")
                    sys.exit()
                available_cpu_cores = os.cpu_count()

############################################################################################################################
                if module_arg == "correction":
                    # try: 
                    if c_lst:
                        config = Config(opts_dict[c_lst[0]], logger) 
                        if i_lst:
                            config.input_file = opts_dict[i_lst[0]]
                        if not os.path.exists(config.input_file):
                            logger.exception('Must set sequencing dataset in configuration.')
                            sys.exit()
                        if t_lst:
                            config.ground_truth_data = opts_dict[t_lst[0]]
                            if not os.path.exists(config.ground_truth_data):
                                logger.error('Ground truth data does not exsit.')
                    elif i_lst:
                        config = Config(None, logger)  
                        config.input_file = opts_dict[i_lst[0]]
                        if not os.path.exists(config.input_file):
                            logger.exception('Input file does not exsit.')
                            sys.exit()
                    else:
                        logger.error('Must input configuration file or sequencing dataset.')
                        sys.exit()
                    ##############################################################
                    if t_lst:
                        config.ground_truth_data = opts_dict[t_lst[0]] 
                        if not os.path.exists(config.ground_truth_data):
                            logger.exception('Ground truth data does not exsit.')
                            raise
                    if d_lst:
                        config.result_dir = opts_dict[d_lst[0]] 
                    if p_lst:
                        config.num_workers = int(opts_dict[p_lst[0]])      
                    if a_lst:
                        config.high_ambiguous = eval(opts_dict[a_lst[0]])
                        # print(config.high_ambiguous)
                    if g_lst:
                        config.tree_method = opts_dict[g_lst[0]]
                    # if o_lst:
                    #     if opts_dict[o_lst[0]] == 'False' or 'false':
                    #         config.over_sampling = False
                    #     elif opts_dict[o_lst[0]] == 'True' or 'true':
                    #         config.over_sampling = True
                    #     else:
                    #         logger.exception("Wrongly set over_sampling")
                    #         raise
                    if config.num_workers <= 0:
                        config.num_workers = available_cpu_cores
                    if config.num_workers > available_cpu_cores:
                        logger.error(f"Only {available_cpu_cores} available to use.") 
                        config.num_workers = available_cpu_cores
            
                    ##############################################################
                    DG = DataGneration(logger, config)
                    if config.high_ambiguous:
                        isolates_file, non_isolates_file, unique_seqs, read_max_len, read_min_len, genuine_df, negative_df, ambiguous_df, high_ambiguous_df = DG.data_files(edit_dis=1)
                    else:
                        isolates_file, non_isolates_file, unique_seqs, read_max_len, read_min_len, genuine_df, negative_df, ambiguous_df = DG.data_files(edit_dis=1)      
                    config.read_max_len = read_max_len
                    MM.measure()
                    ###############################################################
                    EC = ErrorCorrection(logger, config)
                    ## one model to predict

                    if config.high_ambiguous:
                        # corrected_file, no_high_correct, new_negative_df = EC.all_in_one_correction(isolates_file, non_isolates_file, unique_seqs, genuine_df, negative_df, ambiguous_df, high_ambiguous_df)
                        corrected_file, new_negative_df = EC.all_in_one_correction(isolates_file, non_isolates_file, unique_seqs, genuine_df, negative_df, ambiguous_df, high_ambiguous_df)
                    else:
                        corrected_file, new_negative_df = EC.all_in_one_correction(isolates_file, non_isolates_file, unique_seqs, genuine_df, negative_df, ambiguous_df, high_ambiguous_df =None) 
                        
                    if read_min_len > config.min_read_len:
                        genuine_df2, negative_df2, ambiguous_df2, unique_seqs2 = DG.extract_ed2_errors(corrected_file)
                        config.correct_data = EC.all_in_one_ed2_correction(corrected_file, unique_seqs2, genuine_df2, negative_df2, ambiguous_df2)
                    else:
                        config.correct_data = corrected_file
                    del DG, EC
                    MM.measure()
                    ############################
                    DataAnalysis(logger, config).evaluation()  

                    # delete bcool result
                    bcool_dir = os.path.join(config.result_dir, 'bcool/')
                    if os.path.exists(bcool_dir):
                        os.system("rm -rf %s" % bcool_dir)
                    MM.measure()
                    MM.stop()
                    ###############################################################
                    # if config.high_ambiguous:
                    #     if read_min_len > config.min_read_len:
                    #         genuine_df3, negative_df3, ambiguous_df3, unique_seqs3 = DG.extract_ed2_errors(no_high_correct)
                    #         config.correct_data = EC.all_in_one_ed2_correction(no_high_correct, unique_seqs3, genuine_df3, negative_df3, ambiguous_df3)
                    #     else:
                    #         config.correct_data = no_high_correct
                    # DataAnalysis(logger, config).evaluation()   
############################################################################################################################
                elif module_arg == "simplify_correction":
                    # try: 
                    if c_lst:
                        config = Config(opts_dict[c_lst[0]], logger) 
                        if i_lst:
                            config.input_file = opts_dict[i_lst[0]]
                        if not os.path.exists(config.input_file):
                            logger.exception('Must set sequencing dataset in configuration.')
                            sys.exit()
                        if t_lst:
                            config.ground_truth_data = opts_dict[t_lst[0]]
                            if not os.path.exists(config.ground_truth_data):
                                logger.error('Ground truth data does not exsit.')
                    elif i_lst:
                        config = Config(None, logger)  
                        config.input_file = opts_dict[i_lst[0]]
                        if not os.path.exists(config.input_file):
                            logger.exception('Input file does not exsit.')
                            sys.exit()
                    else:
                        logger.error('Must input configuration file or sequencing dataset.')
                        sys.exit()
                    ##############################################################
                    if t_lst:
                        config.ground_truth_data = opts_dict[t_lst[0]] 
                        if not os.path.exists(config.ground_truth_data):
                            logger.exception('Ground truth data does not exsit.')
                            raise
                    if d_lst:
                        config.result_dir = opts_dict[d_lst[0]] 
                    if p_lst:
                        config.num_workers = int(opts_dict[p_lst[0]])      
                    if a_lst:
                        config.high_ambiguous = eval(opts_dict[a_lst[0]])
                        # print(config.high_ambiguous)
                    if config.num_workers <= 0:
                        config.num_workers = available_cpu_cores
                    if config.num_workers > available_cpu_cores:
                        logger.error(f"Only {available_cpu_cores} available to use.") 
                        config.num_workers = available_cpu_cores
                    ##############################################################
                    DG = DataGneration(logger, config)

                    isolates_file, non_isolates_file, read_max_len, read_min_len, genuine_df, ambiguous_df = DG.simplify_data_files(config.input_file, edit_dis=1)      
                    config.read_max_len = read_max_len
                    MM.measure()
                    ###############################################################
                    EC = ErrorCorrection(logger, config)
                    corrected_file = EC.simplify_correction(isolates_file, non_isolates_file, genuine_df, ambiguous_df)
                    
                    # genuine_df, ambiguous_df = DG.simplify_data_files(config.input_file, edit_dis=2) 
                    # config.correct_data = EC.simplify_2nt_correction(config.input_file, genuine_df, ambiguous_df)

                    if read_min_len > config.min_read_len:
                        genuine_df, ambiguous_df = DG.simplify_data_files(corrected_file, edit_dis=2) 
                        config.correct_data = EC.simplify_2nt_correction(corrected_file, genuine_df, ambiguous_df)
                    else:
                        config.correct_data = corrected_file
                    del DG, EC
                    MM.measure()
                    ###################################
                    DataAnalysis(logger, config).evaluation()  
                    # delete bcool result
                    bcool_dir = os.path.join(config.result_dir, 'bcool/')
                    if os.path.exists(bcool_dir):
                        os.system("rm -rf %s" % bcool_dir) 
                    MM.measure()
                    MM.stop()        
############################################################################################################################
                elif module_arg == "amplicon_correction": 
                    if c_lst:
                        config = Config(opts_dict[c_lst[0]], logger) 
                        if i_lst:
                            config.input_file = opts_dict[i_lst[0]]
                        if not os.path.exists(config.input_file):
                            logger.exception('Must set sequencing dataset in configuration.')
                            sys.exit()
                        if t_lst:
                            config.ground_truth_data = opts_dict[t_lst[0]] 
                            if not os.path.exists(config.ground_truth_data):
                                logger.exception('Ground truth data does not exsit.')
                                raise
                    elif i_lst:
                        config = Config(None, logger)  
                        config.input_file = opts_dict[i_lst[0]]
                        if not os.path.exists(config.input_file):
                            logger.exception('Input file does not exsit.')
                            sys.exit()
                    else:
                        logger.error('Must input configuration file or sequencing dataset.')
                        sys.exit()

                    if t_lst:
                        config.ground_truth_data = opts_dict[t_lst[0]] 
                        if not os.path.exists(config.ground_truth_data):
                            logger.exception('Ground truth data does not exsit.')
                            raise
                    if d_lst:
                        config.result_dir = opts_dict[d_lst[0]] 
                    if p_lst:
                        config.num_workers = int(opts_dict[p_lst[0]])    
                    if a_lst:
                        config.high_ambiguous = eval(opts_dict[a_lst[0]])
                    if g_lst:
                        config.tree_method = opts_dict[g_lst[0]] 
                    # if o_lst:
                    #     if opts_dict[o_lst[0]] == 'False' or 'false':
                    #         config.over_sampling = False
                    #     elif opts_dict[o_lst[0]] == 'True' or 'true':
                    #         config.over_sampling = True
                    #     else:
                    #         logger.exception("Wrongly set over_sampling")
                    #         raise
                    if not os.path.exists(config.input_file):
                        logger.exception('Input file does not exsit.')
                        raise
                    if config.num_workers <= 0:
                        config.num_workers = available_cpu_cores
                    if config.num_workers > available_cpu_cores:
                        logger.error(f"Only {available_cpu_cores} available to use.") 
                        config.num_workers = available_cpu_cores
                    ##############################################################
                    DG = DataGneration(logger, config)
                    if config.high_ambiguous:
                        isolates_file, non_isolates_file, unique_seqs, read_max_len, read_min_len, genuine_df, negative_df, ambiguous_df, high_ambiguous_df = DG.data_files(edit_dis=1)
                    else:
                        isolates_file, non_isolates_file, unique_seqs, read_max_len, read_min_len, genuine_df, negative_df, ambiguous_df = DG.data_files(edit_dis=1)      
                    config.read_max_len = read_max_len
                    MM.measure()
                    ###############################################################
                    EC = ErrorCorrection(logger, config)
                    ## one model to predict
                    if config.high_ambiguous:
                        corrected_file, new_negative_df = EC.all_in_one_correction(isolates_file, non_isolates_file, unique_seqs, genuine_df, negative_df, ambiguous_df, high_ambiguous_df)
                    else:
                        corrected_file, new_negative_df = EC.all_in_one_correction(isolates_file, non_isolates_file, unique_seqs, genuine_df, negative_df, ambiguous_df, high_ambiguous_df =None) 
                    if read_min_len > config.min_read_len:
                        genuine_df2, negative_df2, ambiguous_df2, unique_seqs2 = DG.extract_ed2_errors(corrected_file)
                        mid_result = EC.all_in_one_ed2_correction(corrected_file, unique_seqs2, genuine_df2, negative_df2, ambiguous_df2)
                    else:
                        mid_result = corrected_file
                    # mid_result = config.input_file
                    if (not new_negative_df.empty or not negative_df.empty) and not genuine_df.empty:
                        amplicon_df = DG.extract_amplicon_err_samples(mid_result)
                        config.correct_data = EC.correct_amplicon_err(mid_result, genuine_df, negative_df, new_negative_df, amplicon_df, config.amplicon_threshold_proba)
                    else:
                        logger.warning("No genuine or negative samples for amplicon errors prediction!")
                        config.correct_data = mid_result
                    del DG, EC
                    MM.measure()
                    ########################################
                    DataAnalysis(logger, config).evaluation()
                    # delete bcool result
                    bcool_dir = os.path.join(config.result_dir, 'bcool/')
                    if os.path.exists(bcool_dir):
                        os.system("rm -rf %s" % bcool_dir)
                    MM.measure()
                    MM.stop()
############################################################################################################################
                elif module_arg == "umi_correction": 
                    if c_lst:
                        config = Config(opts_dict[c_lst[0]], logger) 
                        if i_lst:
                            config.input_file = opts_dict[i_lst[0]]
                        if not os.path.exists(config.input_file):
                            logger.exception('Must set sequencing dataset in configuration.')
                            sys.exit()
                        if t_lst:
                            config.ground_truth_data = opts_dict[t_lst[0]] 
                            if not os.path.exists(config.ground_truth_data):
                                logger.exception('Ground truth data does not exsit.')
                                raise
                    elif i_lst:
                        config = Config(None, logger)  
                        config.input_file = opts_dict[i_lst[0]]
                        if not os.path.exists(config.input_file):
                            logger.exception('Input file does not exsit.')
                            sys.exit()
                    else:
                        logger.error('Must input configuration file or sequencing dataset.')
                        sys.exit()
                        
                    if t_lst:
                        config.ground_truth_data = opts_dict[t_lst[0]] 
                        if not os.path.exists(config.ground_truth_data):
                            logger.exception('Ground truth data does not exsit.')
                            raise
                    if d_lst:
                        config.result_dir = opts_dict[d_lst[0]] 
                    if p_lst:
                        config.num_workers = int(opts_dict[p_lst[0]])  
                    # if a_lst:
                    #     config.high_ambiguous = eval(opts_dict[a_lst[0]])

                    if config.num_workers <= 0:
                        config.num_workers = available_cpu_cores
                    if config.num_workers > available_cpu_cores:
                        logger.error(f"Only {available_cpu_cores} available to use.") 
                        config.num_workers = available_cpu_cores
                    ##############################################################
                    config.high_ambiguous=False
                    DG = DataGneration(logger, config)
                    genuine_df = DG.extract_umi_genuine_errs(config.input_file)
                    MM.measure()
                    # ##############################################################
                    EC = ErrorCorrection(logger, config)
                    config.correct_data = EC.umi_correction(config.input_file, genuine_df)
                    del DG, EC
                    MM.measure()
                    DataAnalysis(logger, config).evaluation()
                    MM.measure()
                    MM.stop()
############################################################################################################################
                elif module_arg == "mimic_umi":   
                    if i_lst and t_lst:
                        config = Config(None, logger)  
                        config.input_file = opts_dict[i_lst[0]]
                        config.ground_truth_data = opts_dict[t_lst[0]] 
                        if not os.path.exists(config.input_file):
                            logger.exception('Input file does not exsit.')
                            sys.exit()
                        if not os.path.exists(config.ground_truth_data):
                            logger.exception('Ground truth data does not exsit.')
                            sys.exit()
                    else:
                        logger.error("Must input raw and true datasets simultaneously or set them in the configuration file.")
                        sys.exit()

                    if d_lst:
                        config.result_dir = opts_dict[d_lst[0]] 

                    if config.num_workers <= 0:
                        config.num_workers = available_cpu_cores
                    if config.num_workers > available_cpu_cores:
                        logger.error(f"Only {available_cpu_cores} available to use.") 
                        config.num_workers = available_cpu_cores
                #############################################################
                    DP = DataProcessing(logger, config)
                    DP.write_mimic_umis(config.input_file, config.ground_truth_data)
############################################################################################################################
                elif module_arg == "real_umi":   
                    if c_lst:
                        config = Config(opts_dict[c_lst[0]], logger) 
                        if i_lst:
                            config.input_file = opts_dict[i_lst[0]]
                        if not os.path.exists(config.input_file):
                            logger.exception('Must set sequencing dataset in configuration.')
                            sys.exit()
                    elif i_lst:
                        config = Config(None, logger)  
                        config.input_file = opts_dict[i_lst[0]]
                        if not os.path.exists(config.input_file):
                            logger.exception('Input file does not exsit.')
                            sys.exit()
                    else:
                        logger.error("Must input umi-based sequencing dataset or configuration file.")
                        sys.exit()

                    if d_lst:
                        config.result_dir = opts_dict[d_lst[0]] 
                    if config.num_workers <= 0:
                        config.num_workers = available_cpu_cores
                    if config.num_workers > available_cpu_cores:
                        logger.error(f"Only {available_cpu_cores} available to use.") 
                        config.num_workers = available_cpu_cores

                    if u_lst:
                        config.umi_file = opts_dict[u_lst[0]] 
                        if not os.path.exists(config.umi_file):
                            logger.exception('Umi data does not exsit.')
                            sys.exit()
                    DP = DataProcessing(logger, config)
                    if u_lst:
                        if config.umi_file and config.input_file:
                            DP.umi2groundtruth()
                    else:
                        if not config.umi_in_read:
                            DP.real_umi_in_name_data()
                        else:
                            DP.real_umi_data(config.input_file)  
                                              
############################################################################################################################
                elif module_arg == "evaluation":   
                    if i_lst and t_lst and r_lst:
                        config = Config(None, logger)  
                        config.input_file = opts_dict[i_lst[0]]
                        config.ground_truth_data = opts_dict[t_lst[0]]
                        config.correct_data = opts_dict[r_lst[0]]
                        if not os.path.exists(config.input_file):
                            logger.exception('Input file does not exsit.')
                            sys.exit()
                        if not os.path.exists(config.ground_truth_data):
                            logger.exception('Ground truth data does not exsit.')
                            sys.exit()
                        if not os.path.exists(config.correct_data):
                            logger.exception('Corrected data does not exsit.')
                            sys.exit()      

                        if d_lst:
                            config.result_dir = opts_dict[d_lst[0]]
                    elif i_lst and r_lst:
                        config = Config(None, logger)  
                        config.input_file = opts_dict[i_lst[0]]
                        config.correct_data = opts_dict[r_lst[0]]
                        if not os.path.exists(config.input_file):
                            logger.exception('Input file does not exsit.')
                            sys.exit()
                        if not os.path.exists(config.correct_data):
                            logger.exception('Corrected data does not exsit.')
                            sys.exit()   
                        if d_lst:
                            config.result_dir = opts_dict[d_lst[0]]
                    else:
                        logger.error("Must input raw, true and rectified datasets or raw and rectified datasets simultaneously or set them in the configuration file.")
                        sys.exit()

                    if config.num_workers <= 0:
                        config.num_workers = available_cpu_cores
                    if config.num_workers > available_cpu_cores:
                        logger.error(f"Only {available_cpu_cores} available to use.") 
                        config.num_workers = available_cpu_cores
                    if tau_lst:
                        config.high_freq_thre = int(opts_dict[tau_lst[0]])
                #############################################################
                    DataAnalysis(logger, config).evaluation()
############################################################################################################################
                elif module_arg == "simulation":  
                    if c_lst:
                        config = Config(opts_dict[c_lst[0]], logger) 
                        if i_lst:
                            config.input_file = opts_dict[i_lst[0]]
                        if not os.path.exists(config.input_file):
                            logger.exception('Must set sequencing dataset in configuration.')
                            sys.exit()
                    elif i_lst:
                        config = Config(None, logger)  
                        config.input_file = opts_dict[i_lst[0]]
                        if not os.path.exists(config.input_file):
                            logger.exception('Input file does not exsit.')
                            sys.exit()
                    else:
                        logger.error('Must input configuration file or sequencing dataset.')
                        sys.exit()

                    if d_lst:
                        config.result_dir = opts_dict[d_lst[0]] 
                    if a_lst:
                        config.high_ambiguous = eval(opts_dict[a_lst[0]])
                    if g_lst:
                        config.tree_method = opts_dict[g_lst[0]]
                    # if o_lst:
                    #     if opts_dict[o_lst[0]] == 'False' or 'false':
                    #         config.over_sampling = False
                    #     elif opts_dict[o_lst[0]] == 'True' or 'true':
                    #         config.over_sampling = True
                    #     else:
                    #         logger.exception("Wrongly set over_sampling")
                    #         raise
                    if config.num_workers <= 0:
                        config.num_workers = available_cpu_cores
                    if config.num_workers > available_cpu_cores:
                        logger.error(f"Only {available_cpu_cores} available to use.") 
                        config.num_workers = available_cpu_cores
                    Simulation(logger, config).simulation()
                    # delete bcool result
                    bcool_dir = os.path.join(config.result_dir, 'bcool/')
                    if os.path.exists(bcool_dir):
                        os.system("rm -rf %s" % bcool_dir)
############################################################################################################################
                # elif module_arg == "extract_isolates": 

                else:
                    # logger.error("Invalid module name, please check.")
                    logger.error("Input wrong module name, using command noise2read -h/--help for usage.")
                    sys.exit()
        
        else:
            logger.error("No valid arguments input, using command noise2read -h/--help for usage")
    except getopt.GetoptError as err:
        logger.error(err)
